#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import errno, os, time


def try_for(fn, timeout=5.0, delay=0.1):
    """Calls ``fn`` every ``delay`` seconds until it returns True or ``timeout``
    seconds elapse. Returns True if the lock was acquired, or False if the timeout
    was reached.

    :param timeout: Length of time (in seconds) to keep retrying to acquire the
        lock. 0 means return immediately. Only used when blocking is False.
    :param delay: How often (in seconds) to retry acquiring the lock during
        the timeout period. Only used when blocking is False and timeout > 0.
    """
    
    until = time.time() + timeout
    v = fn()
    while not v and time.time() < until:
        time.sleep(delay)
        v = fn()
    return v


class LockBase(object):
    """Base class for file locks.
    """
    
    def __init__(self, filename):
        self.fd = None
        self.filename = filename
        self.locked = False
    
    def __del__(self):
        if hasattr(self, "fd") and self.fd:
            try:
                self.release()
            except:
                pass
    
    def acquire(self, blocking=False):
        """Acquire the lock. Returns True if the lock was acquired.
        
        :param blocking: if True, call blocks until the lock is acquired.
            This may not be available on all platforms. On Windows, this is
            actually just a delay of 10 seconds, rechecking every second.
        """
        pass
    
    def release(self):
        pass
    
    
class FcntlLock(LockBase):
    """File lock based on UNIX-only fcntl module.
    """
    
    def acquire(self, blocking=False):
        import fcntl
        
        flags = os.O_CREAT | os.O_WRONLY
        self.fd = os.open(self.filename, flags)
        
        mode = fcntl.LOCK_EX
        if not blocking: mode |= fcntl.LOCK_NB
        
        try:
            fcntl.flock(self.fd, mode)
            self.locked = True
            return True
        except IOError, e:
            if e.errno not in (errno.EAGAIN, errno.EACCES):
                raise
            os.close(self.fd)
            self.fd = None
            return False
        
    def release(self):
        import fcntl
        fcntl.flock(self.fd, fcntl.LOCK_UN)
        os.close(self.fd)
        self.fd = None


class MsvcrtLock(LockBase):
    """File lock based on Windows-only msvcrt module.
    """
    
    def acquire(self, blocking=False):
        import msvcrt
        
        flags = os.O_CREAT | os.O_WRONLY
        mode = msvcrt.LK_NBLCK
        if blocking: mode = msvcrt.LK_LOCK
            
        self.fd = os.open(self.filename, flags)
        try:
            msvcrt.locking(self.fd, mode, 1)
            return True
        except IOError, e:
            if e.errno not in (errno.EAGAIN, errno.EACCES, errno.EDEADLK):
                raise
            os.close(self.fd)
            self.fd = None
            return False
        
    def release(self):
        if self.fd is None:
            raise Exception("Lock was not acquired")
        
        import msvcrt
        msvcrt.locking(self.fd, msvcrt.LK_UNLCK, 1)
        os.close(self.fd)
        self.fd = None


if os.name == "nt":
    FileLock = MsvcrtLock
else:
    FileLock = FcntlLock






import unittest
import os.path
from random import random, randint

from whoosh.formats import (Characters, CharacterBoosts, DocBoosts, Existence,
                            Frequency, Positions, PositionBoosts)
from whoosh.filedb.filestore import FileStorage
from whoosh.filedb.filepostings import FilePostingWriter, FilePostingReader
from whoosh.util import float_to_byte, byte_to_float


class TestPostings(unittest.TestCase):
    def make_postings(self):
        postings = [(1, 23), (3, 45), (12, 2), (34, 21), (43, 7), (67, 103), (68, 1), (102, 31),
                    (145, 4), (212, 9), (283, 30), (291, 6), (412, 39), (900, 50), (905, 28), (1024, 8),
                    (1800, 13), (2048, 3), (15000, 40)]
        return postings
    
    def make_file(self, name):
        if not os.path.exists("testindex"):
            os.mkdir("testindex")
        return FileStorage("testindex").create_file(name+"_test.pst")
    
    def open_file(self, name):
        return FileStorage("testindex").open_file(name+"_test.pst")
    
    def delete_file(self, name):
        try:
            FileStorage("testindex").delete_file(name+"_test.pst")
        except OSError:
            raise
    
    def test_readwrite(self):
        format = Frequency(None)
        postings = self.make_postings()
        
        postfile = self.make_file("readwrite")
        fpw = FilePostingWriter(postfile, blocklimit=8)
        fpw.start(format)
        for id, freq in postings:
            fpw.write(id, float(freq), format.encode(freq), 0)
        fpw.close()
        
        postfile = self.open_file("readwrite")
        fpr = FilePostingReader(postfile, 0, format)
        self.assertEqual(postings, list(fpr.items_as("frequency")))
        postfile.close()
        self.delete_file("readwrite")
        
    def test_skip(self):
        format = Frequency(None)
        postings = self.make_postings()
        
        postfile = self.make_file("skip")
        fpw = FilePostingWriter(postfile, blocklimit=8)
        fpw.start(format)
        for id, freq in postings:
            fpw.write(id, float(freq), format.encode(freq), 0)
        fpw.close()
        
        postfile = self.open_file("skip")
        fpr = FilePostingReader(postfile, 0, format)
        fpr.skip_to(220)
        self.assertEqual(postings[10:], list(fpr.items_as("frequency")))
        postfile.close()
        self.delete_file("skip")
    
    def roundtrip(self, postings, format, astype):
        postfile = self.make_file(astype)
        getweight = format.decoder("weight")
        fpw = FilePostingWriter(postfile, blocklimit=8)
        fpw.start(format)
        for id, value in postings:
            v = format.encode(value)
            fpw.write(id, getweight(v), v, 0)
        fpw.close()
        
        postfile = self.open_file(astype)
        fpr = FilePostingReader(postfile, 0, format)
        readback = list(fpr.items_as(astype))
        postfile.close()
        self.delete_file(astype)
        return readback
    
    def test_existence_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            postings.append((docnum, 1))
        
        self.assertEqual(postings, self.roundtrip(postings, Existence(None), "frequency"))
    
    def test_docboost_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            freq = randint(1, 1000)
            boost = byte_to_float(float_to_byte(random() * 2))
            postings.append((docnum, (freq, boost)))
        
        self.assertEqual(postings, self.roundtrip(postings, DocBoosts(None), "docboosts"))
        
    def test_position_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            for __ in xrange(0, randint(1, 10)):
                pos += randint(1, 10)
                posns.append(pos)
            postings.append((docnum, posns))
        
        self.assertEqual(postings, self.roundtrip(postings, Positions(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in postings]
        self.assertEqual(as_freq, self.roundtrip(postings, Positions(None), "frequency"))
        
    def test_character_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            endchar = 0
            for __ in xrange(0, randint(1, 10)):
                pos += randint(1, 10)
                startchar = endchar + randint(3, 10)
                endchar = startchar + randint(3, 10)
                posns.append((pos, startchar, endchar))
            postings.append((docnum, posns))
            
        self.assertEqual(postings, self.roundtrip(postings, Characters(None), "characters"))
        
        as_posns = [(docnum, [pos for pos, sc, ec in posns]) for docnum, posns in postings]
        self.assertEqual(as_posns, self.roundtrip(postings, Characters(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in as_posns]
        self.assertEqual(as_freq, self.roundtrip(postings, Characters(None), "frequency"))
        
    def test_posboost_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 3):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            for __ in xrange(0, randint(1, 3)):
                pos += randint(1, 10)
                boost = byte_to_float(float_to_byte(random() * 2))
                posns.append((pos, boost))
            postings.append((docnum, posns))
        
        self.assertEqual(postings, self.roundtrip(postings, PositionBoosts(None), "position_boosts"))
        
        as_posns = [(docnum, [pos for pos, boost in posns]) for docnum, posns in postings]
        self.assertEqual(as_posns, self.roundtrip(postings, PositionBoosts(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in postings]
        self.assertEqual(as_freq, self.roundtrip(postings, PositionBoosts(None), "frequency"))

    def test_charboost_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            endchar = 0
            for __ in xrange(0, randint(1, 10)):
                pos += randint(1, 10)
                startchar = endchar + randint(3, 10)
                endchar = startchar + randint(3, 10)
                boost = byte_to_float(float_to_byte(random() * 2))
                posns.append((pos, startchar, endchar, boost))
            postings.append((docnum, posns))
        
        self.assertEqual(postings, self.roundtrip(postings, CharacterBoosts(None), "character_boosts"))
        
        as_chars = [(docnum, [(pos, sc, ec) for pos, sc, ec, bst in posns]) for docnum, posns in postings]
        self.assertEqual(as_chars, self.roundtrip(postings, CharacterBoosts(None), "characters"))
        
        as_posbsts = [(docnum, [(pos, bst) for pos, sc, ec, bst in posns]) for docnum, posns in postings]
        self.assertEqual(as_posbsts, self.roundtrip(postings, CharacterBoosts(None), "position_boosts"))
        
        as_posns = [(docnum, [pos for pos, sc, ec, bst in posns]) for docnum, posns in postings]
        self.assertEqual(as_posns, self.roundtrip(postings, CharacterBoosts(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in as_posns]
        self.assertEqual(as_freq, self.roundtrip(postings, CharacterBoosts(None), "frequency"))

    
        
        


if __name__ == '__main__':
    unittest.main()





import unittest

import os.path
from shutil import rmtree

from whoosh import query
from whoosh.fields import *
from whoosh.ramdb.ramindex import RamIndex


class TestRamIndex(unittest.TestCase):
    def make_index(self):
        ana = StandardAnalyzer(stoplist=None)
        sc = Schema(id=ID(stored=True),
                    text=TEXT(analyzer=ana, vector=Frequency(ana)),
                    subs=NUMERIC(int, stored=True))
        ix = RamIndex(sc)
        
        ix.add_document(id=u"fieldtype",
                        text=u"The FieldType object supports the following attributes",
                        subs=56)
        ix.add_document(id=u"format",
                        text=u"the storage format for the field contents",
                        subs=100)
        ix.add_document(id=u"vector",
                        text=u"the storage format for the field vectors (forward index)",
                        subs=23)
        ix.add_document(id=u"scorable",
                        text=u"whether searches against this field may be scored.",
                        subs=34)
        ix.add_document(id=u"stored",
                        text=u"whether the content of this field is stored for each document.",
                        subs=575)
        ix.add_document(id=u"unique",
                        text=u"whether this field value is unique to each document.",
                        subs=2)
        ix.add_document(id=u"const",
                        text=u"The constructor for the base field type simply",
                        subs=58204)
        return ix
    
    def test_indexing(self):
        ix = self.make_index()
        s = ix.searcher()
        q = query.Term("text", "format")
        r = s.search(q)
        self.assertEqual(len(r), 2)
        self.assertEqual(r[0]["id"], "format")
        self.assertEqual(r[0]["subs"], 100)
        self.assertEqual(r[1]["id"], "vector")
        self.assertEqual(r[1]["subs"], 23)

    def test_deleting(self):
        ix = self.make_index()
        ix.delete_by_term("id", u"vector")
        
        s = ix.searcher()
        q = query.Term("text", "format")
        r = s.search(q)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0]["id"], "format")
        self.assertEqual(r[0]["subs"], 100)
        
    def test_optimize(self):
        ix = self.make_index()
        self.assertEqual(ix.doc_count(), 7)
        self.assertEqual(ix.doc_count_all(), 7)
        
        ix.delete_by_term("id", u"vector")
        self.assertEqual(ix.doc_count(), 6)
        self.assertEqual(ix.doc_count_all(), 7)
        self.assertTrue(ix.has_deletions())
        
        ix.optimize()
        self.assertEqual(ix.doc_count(), 6)
        self.assertEqual(ix.doc_count_all(), 6)
        self.assertFalse(ix.has_deletions())
        self.assertFalse(("text", u"forward") in ix.reader())
        
        ix.close()
        
    def test_reader(self):
        ix = self.make_index()
        r = ix.reader()
        
        self.assertTrue(("text", u"constructor") in r)
        self.assertEqual(r.stored_fields(2), {"id": "vector", "subs": 23})
        
        target = [{"id": u"fieldtype", "subs": 56},
                  {"id": u"format", "subs": 100},
                  {"id": u"vector", "subs": 23},
                  {"id": u"scorable", "subs": 34},
                  {"id": u"stored", "subs": 575},
                  {"id": u"unique", "subs": 2},
                  {"id": u"const", "subs": 58204},
                  ]
        self.assertEqual(list(r.all_stored_fields()), target)
        
        self.assertEqual(r.field_length("text"), 59)
        self.assertEqual(r.max_field_length("text"), 11)
        self.assertEqual(r.doc_field_length(3, "text"), 8)
        
        self.assertEqual(r.doc_frequency("text", "the"), 5)
        self.assertEqual(r.frequency("text", "the"), 9)
        
        everything = [("id", u'const', 1, 1), ("id", u'fieldtype', 1, 1), ("id", u'format', 1, 1),
                      ("id", u'scorable', 1, 1), ("id", u'stored', 1, 1), ("id", u'unique', 1, 1),
                      ("id", u'vector', 1, 1),  ("text", u'against', 1, 1),
                      ("text", u'attributes', 1, 1), ("text", u'base', 1, 1), ("text", u'be', 1, 1),
                      ("text", u'constructor', 1, 1), ("text", u'content', 1, 1), ("text", u'contents', 1, 1),
                      ("text", u'document', 2, 2), ("text", u'each', 2, 2), ("text", u'field', 6, 6),
                      ("text", u'fieldtype', 1, 1), ("text", u'following', 1, 1), ("text", u'for', 4, 4),
                      ("text", u'format', 2, 2), ("text", u'forward', 1, 1), ("text", u'index', 1, 1),
                      ("text", u'is', 2, 2), ("text", u'may', 1, 1), ("text", u'object', 1, 1),
                      ("text", u'of', 1, 1), ("text", u'scored', 1, 1), ("text", u'searches', 1, 1),
                      ("text", u'simply', 1, 1), ("text", u'storage', 2, 2), ("text", u'stored', 1, 1),
                      ("text", u'supports', 1, 1), ("text", u'the', 5, 9), ("text", u'this', 3, 3),
                      ("text", u'to', 1, 1), ("text", u'type', 1, 1), ("text", u'unique', 1, 1),
                      ("text", u'value', 1, 1), ("text", u'vectors', 1, 1), ("text", u'whether', 3, 3)]
        
        self.assertEqual([item for item in r if item[0] != 'subs'], everything)
        self.assertEqual(list(r.iter_from("text", u"su")), everything[32:])
        
    def test_vectors(self):
        ix = self.make_index()
        r = ix.reader()
        
        self.assertFalse(r.has_vector(0, "id"))
        self.assertTrue(r.has_vector(0, "text"))
        
        target = [(u'contents', 1), (u'field', 1), (u'for', 1), (u'format', 1),
                  (u'storage', 1), (u'the', 2)]
        vec = list(r.vector_as("frequency", 1, "text"))
        self.assertEqual(target, vec)
        
    def test_todisk(self):
        if not os.path.exists("testindex"):
            os.mkdir("testindex")

        ix = self.make_index()

        from whoosh.index import create_in
        fix = create_in("testindex", ix.schema)
        
        w = fix.writer()
        w.add_reader(ix.reader())
        w.commit()
        
        rmtree("testindex", ignore_errors=True)


if __name__ == '__main__':
    unittest.main()




#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""Contains the main functions/classes for creating, maintaining, and using
an index.
"""

from __future__ import division
import os.path

from whoosh import fields, store


_DEF_INDEX_NAME = "MAIN"


# Exceptions

class IndexError(Exception):
    """Generic index error."""

class IndexVersionError(IndexError):
    """Raised when you try to open an index using a format that the current
    version of Whoosh cannot read. That is, when the index you're trying to
    open is either not backward or forward compatible with this version of
    Whoosh.
    """
    
    def __init__(self, msg, version, release=None):
        Exception.__init__(self, msg)
        self.version = version
        self.release = release

class OutOfDateError(IndexError):
    """Raised when you try to commit changes to an index which is not the
    latest generation.
    """

class EmptyIndexError(IndexError):
    """Raised when you try to work with an index that has no indexed terms.
    """


# Convenience functions

def create_in(dirname, schema, indexname=None):
    """Convenience function to create an index in a directory. Takes care of
    creating a FileStorage object for you.
    
    :param dirname: the path string of the directory in which to create the index.
    :param schema: a :class:`whoosh.fields.Schema` object describing the index's fields.
    :param indexname: the name of the index to create; you only need to specify this if
        you are creating multiple indexes within the same storage object.
    :returns: :class:`Index`
    """
    
    if not indexname:
        indexname = _DEF_INDEX_NAME
    
    from whoosh.filedb.filestore import FileStorage
    storage = FileStorage(dirname)
    return storage.create_index(schema, indexname)

def open_dir(dirname, indexname=None, mapped=True):
    """Convenience function for opening an index in a directory. Takes care of
    creating a FileStorage object for you. dirname is the filename of the
    directory in containing the index. indexname is the name of the index to
    create; you only need to specify this if you have multiple indexes within
    the same storage object.
    
    :param dirname: the path string of the directory in which to create the
        index.
    :param indexname: the name of the index to create; you only need to specify
        this if you have multiple indexes within the same storage object.
    :param mapped: whether to use memory mapping to speed up disk reading.
    :returns: :class:`Index`
    """
    
    if indexname is None:
        indexname = _DEF_INDEX_NAME
    
    from whoosh.filedb.filestore import FileStorage
    storage = FileStorage(dirname, mapped=mapped)
    return storage.open_index(indexname)

def exists_in(dirname, indexname=None):
    """Returns True if dirname contains a Whoosh index.
    
    :param dirname: the file path of a directory.
    :param indexname: the name of the index. If None, the default index name is
        used.
    :param rtype: bool
    """
    
    if os.path.exists(dirname):
        try:
            ix = open_dir(dirname, indexname=indexname)
            return ix.latest_generation() > -1
        except EmptyIndexError:
            pass

    return False

def exists(storage, indexname=None):
    """Returns True if the given Storage object contains a Whoosh index.
    
    :param storage: a store.Storage object.
    :param indexname: the name of the index. If None, the default index name is
        used.
    :param rtype: bool
    """
    
    if indexname is None:
        indexname = _DEF_INDEX_NAME
        
    try:
        ix = storage.open_index(indexname)
        gen = ix.latest_generation()
        ix.close()
        return gen > -1
    except EmptyIndexError:
        pass
    
    return False

def version_in(dirname, indexname=None):
    """Returns a tuple of (release_version, format_version), where
    release_version is the release version number of the Whoosh code that
    created the index -- e.g. (0, 1, 24) -- and format_version is the version
    number of the on-disk format used for the index -- e.g. -102.
    
    The second number (format version) may be useful for figuring out if you
    need to recreate an index because the format has changed. However, you can
    just try to open the index and see if you get an IndexVersionError
    exception.
    
    Note that the release and format version are available as attributes on the
    Index object in Index.release and Index.version.
    
    :param dirname: the file path of a directory containing an index.
    :param indexname: the name of the index. If None, the default index name is
        used.
    :returns: ((major_ver, minor_ver, build_ver), format_ver)
    """
    
    from whoosh.filedb.filestore import FileStorage
    storage = FileStorage(dirname)
    return version(storage, indexname=indexname)
    

def version(storage, indexname=None):
    """Returns a tuple of (release_version, format_version), where
    release_version is the release version number of the Whoosh code that
    created the index -- e.g. (0, 1, 24) -- and format_version is the version
    number of the on-disk format used for the index -- e.g. -102.
    
    The second number (format version) may be useful for figuring out if you
    need to recreate an index because the format has changed. However, you can
    just try to open the index and see if you get an IndexVersionError
    exception.
    
    Note that the release and format version are available as attributes on the
    Index object in Index.release and Index.version.
    
    :param storage: a store.Storage object.
    :param indexname: the name of the index. If None, the default index name is
        used.
    :returns: ((major_ver, minor_ver, build_ver), format_ver)
    """
    
    try:
        if indexname is None:
            indexname = _DEF_INDEX_NAME
        
        ix = storage.open_index(indexname)
        return (ix.release, ix.version)
    except IndexVersionError, e:
        return (None, e.version)


# Index class

class Index(object):
    """Represents an indexed collection of documents.
    """
    
    def close(self):
        """Closes any open resources held by the Index object itself. This may
        not close all resources being used everywhere, for example by a
        Searcher object.
        """
        pass
    
    def add_field(self, fieldname, fieldspec):
        """Adds a field to the index's schema.
        
        :param fieldname: the name of the field to add.
        :param fieldspec: an instantiated :class:`whoosh.fields.FieldType`
            object.
        """
        
        w = self.writer()
        w.add_field(fieldname, fieldspec)
        w.commit()
        
    def remove_field(self, fieldname):
        """Removes the named field from the index's schema. Depending on the
        backend implementation, this may or may not actually remove existing
        data for the field from the index. Optimizing the index should always
        clear out existing data for a removed field.
        """
        
        w = self.writer()
        w.remove_field(fieldname)
        w.commit()
        
    def latest_generation(self):
        """Returns the generation number of the latest generation of this
        index, or -1 if the backend doesn't support versioning.
        """
        return -1
    
    def refresh(self):
        """Returns a new Index object representing the latest generation
        of this index (if this object is the latest generation, or the backend
        doesn't support versioning, returns self).
        
        :returns: :class:`Index`
        """
        return self
    
    def up_to_date(self):
        """Returns True if this object represents the latest generation of
        this index. Returns False if this object is not the latest generation
        (that is, someone else has updated the index since you opened this
        object).
        
        :param rtype: bool
        """
        return True
    
    def last_modified(self):
        """Returns the last modified time of the index, or -1 if the backend
        doesn't support last-modified times.
        """
        return - 1
    
    def is_empty(self):
        """Returns True if this index is empty (that is, it has never had any
        documents successfully written to it.
        
        :param rtype: bool
        """
        raise NotImplementedError
    
    def optimize(self):
        """Optimizes this index, if necessary.
        """
        pass
    
    def doc_count_all(self):
        """Returns the total number of documents, DELETED OR UNDELETED,
        in this index.
        """
        
        r = self.reader()
        try:
            return r.doc_count_all()
        finally:
            r.close()
    
    def doc_count(self):
        """Returns the total number of UNDELETED documents in this index.
        """
        
        r = self.reader()
        try:
            return r.doc_count()
        finally:
            r.close()
    
    def searcher(self, **kwargs):
        """Returns a Searcher object for this index. Keyword arguments are
        passed to the Searcher object's constructor.
        
        :rtype: :class:`whoosh.searching.Searcher`
        """
        
        from whoosh.searching import Searcher
        return Searcher(self, **kwargs)
    
    def field_length(self, fieldname):
        """Returns the total length of the field across all documents.
        """
        
        r = self.reader()
        try:
            return r.field_length(fieldname)
        finally:
            r.close()
    
    def max_field_length(self, fieldname):
        """Returns the maximum length of the field across all documents.
        """
        
        r = self.reader()
        try:
            return r.max_field_length(fieldname)
        finally:
            r.close()
    
    def reader(self):
        """Returns an IndexReader object for this index.
        
        :rtype: :class:`whoosh.reading.IndexReader`
        """
        raise NotImplementedError
    
    def writer(self, **kwargs):
        """Returns an IndexWriter object for this index.
        
        :rtype: :class:`whoosh.writing.IndexWriter`
        """
        raise NotImplementedError
    
    def delete_by_term(self, fieldname, text, searcher=None):
        w = self.writer()
        w.delete_by_term(fieldname, text, searcher=searcher)
        w.commit()
        
    def delete_by_query(self, q, searcher=None):
        w = self.writer()
        w.delete_by_query(q, searcher=searcher)
        w.commit()
        
    

# Debugging functions

        
if __name__ == '__main__':
    pass



import os.path, bz2, xml.sax.handler

from whoosh.util import now


#<mediawiki xml:lang="en">
#   <page>
#     <title>Page title</title>
#     <restrictions>edit=sysop:move=sysop</restrictions>
#     <revision>
#       <timestamp>2001-01-15T13:15:00Z</timestamp>
#       <contributor><username>Foobar</username></contributor>
#       <comment>I have just one thing to say!</comment>
#       <text>A bunch of [[text]] here.</text>
#       <minor />
#     </revision>
#     <revision>
#       <timestamp>2001-01-15T13:10:27Z</timestamp>
#       <contributor><ip>10.0.0.2</ip></contributor>
#       <comment>new!</comment>
#       <text>An earlier [[revision]].</text>
#     </revision>
#   </page>
#
#   <page>
#     <title>Talk:Page title</title>
#     <revision>
#       <timestamp>2001-01-15T14:03:00Z</timestamp>
#       <contributor><ip>10.0.0.2</ip></contributor>
#       <comment>hey</comment>
#       <text>WHYD YOU LOCK PAGE??!!! i was editing that jerk</text>
#     </revision>
#   </page>
#</mediawiki>


filename = "C:\Documents and Settings\matt\Desktop\Search\enwiki-latest-pages-meta-current.xml.bz2"
f = bz2.BZ2File(filename, "r")

class WPHandler(xml.sax.handler.ContentHandler):
    def __init__(self):
        self.inpage = False
        self.pagecount = 0
        self.intitle = False
        self.intext = False
        self.textcount = 0
        self.stime = now()
        self.stime_block = now()
        
    def startElement(self, name, attrs):
        if name == "page":
            self.inpage = True
        elif name == "title":
            self.intitle = True
        elif name == "text":
            self.intext = True
        
    def endElement(self, name):
        if name == "page":
            self.inpage = False
            self.pagecount += 1
            if not self.pagecount % 1000:
                n = now()
                t = n - self.stime
                print self.pagecount, self.textcount, n - self.stime_block, t/60
                self.stime_block = n
        elif name == "title":
            self.intitle = False
        elif name == "text":
            self.intext = False
            
    def characters(self, text):
        if self.intitle:
            self.title = text
        elif self.intext:
            self.text = text
            self.textcount += len(text)


t = now()
h = WPHandler()
parser = xml.sax.parse(f, h)
print now() - t
print h.pagecount




import unittest

from whoosh import index, spelling
from whoosh.filedb.filestore import RamStorage


class TestSpelling(unittest.TestCase):
    def test_spelling(self):
        st = RamStorage()
        
        sp = spelling.SpellChecker(st, mingram=2)
        
        wordlist = ["render", "animation", "animate", "shader",
                    "shading", "zebra", "koala", "lamppost",
                    "ready", "kismet", "reaction", "page",
                    "delete", "quick", "brown", "fox", "jumped",
                    "over", "lazy", "dog", "wicked", "erase",
                    "red", "team", "yellow", "under", "interest",
                    "open", "print", "acrid", "sear", "deaf",
                    "feed", "grow", "heal", "jolly", "kilt",
                    "low", "zone", "xylophone", "crown",
                    "vale", "brown", "neat", "meat", "reduction",
                    "blunder", "preaction"]
        
        sp.add_words([unicode(w) for w in wordlist])
        
        sugs = sp.suggest(u"reoction")
        self.assertNotEqual(len(sugs), 0)
        self.assertEqual(sugs, [u"reaction", u"reduction", u"preaction"])
        
    def test_suggestionsandscores(self):
        st = RamStorage()
        sp = spelling.SpellChecker(st, mingram=2)
        
        words = [("alfa", 10), ("bravo", 9), ("charlie", 8), ("delta", 7),
                 ("echo", 6), ("foxtrot", 5), ("golf", 4), ("hotel", 3),
                 ("india", 2), ("juliet", 1)]
        sp.add_scored_words((unicode(w), s) for w, s in words)
        
        from whoosh.scoring import Frequency
        sugs = sp.suggestions_and_scores(u"alpha", weighting=Frequency())
        self.assertEqual(sugs, [(u"alfa", 10, 3.0), (u"charlie", 8, 1.0)])

    def test_minscore(self):
        st = RamStorage()
        sp = spelling.SpellChecker(st, mingram=2, minscore=2.0)
        
        sp.add_words([u'charm', u'amour'])
        
        sugs = sp.suggest(u"armor")
        self.assertEqual(sugs, [u'charm'])



if __name__ == '__main__':
    unittest.main()




import unittest

from whoosh import analysis, fields, formats, qparser, query, searching
from whoosh.filedb.filestore import RamStorage
from whoosh.util import permutations


class TestResults(unittest.TestCase):
    def test_score_retrieval(self):
        schema = fields.Schema(title=fields.TEXT(stored=True),
                               content=fields.TEXT(stored=True))
        storage = RamStorage()
        ix = storage.create_index(schema)
        writer = ix.writer()
        writer.add_document(title=u"Miss Mary",
                            content=u"Mary had a little white lamb its fleece was white as snow")
        writer.add_document(title=u"Snow White",
                            content=u"Snow white lived in the forest with seven dwarfs")
        writer.commit()
        
        searcher = ix.searcher()
        results = searcher.search(query.Term("content", "white"))
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]['title'], u"Miss Mary")
        self.assertEqual(results[1]['title'], u"Snow White")
        self.assertNotEqual(results.score(0), None)
        self.assertNotEqual(results.score(0), 0)
        self.assertNotEqual(results.score(0), 1)

    def test_resultcopy(self):
        schema = fields.Schema(a=fields.TEXT(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(a=u"alfa bravo charlie")
        w.add_document(a=u"bravo charlie delta")
        w.add_document(a=u"charlie delta echo")
        w.add_document(a=u"delta echo foxtrot")
        w.commit()
        
        s = ix.searcher()
        r = s.search(qparser.QueryParser("a").parse(u"charlie"))
        self.assertEqual(len(r), 3)
        rcopy = r.copy()
        self.assertEqual(r.top_n, rcopy.top_n)
        self.assertEqual(r.scores, rcopy.scores)
        
    def test_resultslength(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               value=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"1", value=u"alfa alfa alfa alfa alfa")
        w.add_document(id=u"2", value=u"alfa alfa alfa alfa")
        w.add_document(id=u"3", value=u"alfa alfa alfa")
        w.add_document(id=u"4", value=u"alfa alfa")
        w.add_document(id=u"5", value=u"alfa")
        w.add_document(id=u"6", value=u"bravo")
        w.commit()
        
        s = ix.searcher()
        r = s.search(query.Term("value", u"alfa"), limit=3)
        self.assertEqual(len(r), 5)
        self.assertEqual(r.scored, 3)

    def test_pages(self):
        from whoosh.scoring import Frequency
        
        schema = fields.Schema(id=fields.ID(stored=True), c=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"1", c=u"alfa alfa alfa alfa alfa alfa")
        w.add_document(id=u"2", c=u"alfa alfa alfa alfa alfa")
        w.add_document(id=u"3", c=u"alfa alfa alfa alfa")
        w.add_document(id=u"4", c=u"alfa alfa alfa")
        w.add_document(id=u"5", c=u"alfa alfa")
        w.add_document(id=u"6", c=u"alfa")
        w.commit()
        
        s = ix.searcher(weighting=Frequency)
        q = query.Term("c", u"alfa")
        r = s.search(q)
        self.assertEqual([d["id"] for d in r], ["1", "2", "3", "4", "5", "6"])
        r = s.search_page(q, 2, pagelen=2)
        self.assertEqual([d["id"] for d in r], ["3", "4"])
        
        r = s.search_page(q, 2, pagelen=4)
        self.assertEqual(r.total, 6)
        self.assertEqual(r.pagenum, 2)
        self.assertEqual(r.pagelen, 2)
    
    def test_page_counts(self):
        from whoosh.scoring import Frequency
        
        schema = fields.Schema(id=fields.ID(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        for i in xrange(10):
            w.add_document(id=unicode(i))
        w.commit()
        s = ix.searcher(weighting=Frequency)
        q = query.Every("id")
        
        r = s.search(q)
        self.assertEqual(len(r), 10)
        
        self.assertRaises(ValueError, s.search_page, q, 0)
        
        r = s.search_page(q, 1, 5)
        self.assertEqual(len(r), 10)
        self.assertEqual(r.pagecount, 2)
        
        r = s.search_page(q, 1, 5)
        self.assertEqual(len(r), 10)
        self.assertEqual(r.pagecount, 2)
        
        r = s.search_page(q, 2, 5)
        self.assertEqual(len(r), 10)
        self.assertEqual(r.pagecount, 2)
        self.assertEqual(r.pagenum, 2)
        
        r = s.search_page(q, 1, 10)
        self.assertEqual(len(r), 10)
        self.assertEqual(r.pagecount, 1)
        self.assertEqual(r.pagenum, 1)
    
    def test_resultspage(self):
        schema = fields.Schema(id=fields.STORED, content=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(schema)
        
        domain = ("alfa", "bravo", "bravo", "charlie", "delta")
        w = ix.writer()
        i = 0
        for lst in permutations(domain, 3):
            w.add_document(id=unicode(i), content=u" ".join(lst))
            i += 1
        w.commit()
        
        s = ix.searcher()
        q = query.Term("content", u"bravo")
        r = s.search(q, limit=10)
        tops = list(r)
        
        rp = s.search_page(q, 1, pagelen=5)
        self.assertEqual(list(rp), tops[0:5])
        
        rp = s.search_page(q, 2, pagelen=5)
        self.assertEqual(list(rp), tops[5:10])
        
        rp = s.search_page(q, 1, pagelen=10)
        self.assertEqual(len(rp), 54)
        self.assertEqual(rp.pagecount, 6)
        rp = s.search_page(q, 6, pagelen=10)
        self.assertEqual(len(list(rp)), 4)
        self.assertTrue(rp.is_last_page())
        
        self.assertRaises(ValueError, s.search_page, q, 0)
        self.assertRaises(ValueError, s.search_page, q, 7)
        
        rp = s.search_page(query.Term("content", "glonk"), 1)
        self.assertEqual(len(rp), 0)
        self.assertTrue(rp.is_last_page())
    
    def test_keyterms(self):
        ana = analysis.StandardAnalyzer()
        vectorformat = formats.Frequency(ana)
        schema = fields.Schema(path=fields.ID,
                               content=fields.TEXT(analyzer=ana,
                                                   vector=vectorformat))
        st = RamStorage()
        ix = st.create_index(schema)
        w = ix.writer()
        w.add_document(path=u"a",content=u"This is some generic content")
        w.add_document(path=u"b",content=u"This is some distinctive content")
        w.commit()
        
        s = ix.searcher()
        docnum = s.document_number(path=u"b")
        keyterms = list(s.key_terms([docnum], "content"))
        self.assertTrue(len(keyterms) > 0)
        self.assertEqual(keyterms[0][0], "distinctive")
        
        r = s.search(query.Term("path", u"b"))
        keyterms2 = list(r.key_terms("content"))
        self.assertTrue(len(keyterms2) > 0)
        self.assertEqual(keyterms2[0][0], "distinctive")
        
    def test_facets(self):
        import random
        from collections import defaultdict
        
        schema = fields.Schema(id=fields.STORED, chapter=fields.ID)
        ix = RamStorage().create_index(schema)
        
        w = ix.writer()
        domain = u"alfa bravo charlie delta echo".split()
        targets = dict(zip(domain, (5, 10, 7, 2, 9)))
        cats = defaultdict(list)
        i = 0
        while targets:
            choice = random.choice(targets.keys())
            w.add_document(id=i, chapter=choice)
            cats[choice].append((i, 1.0))
            i += 1
            targets[choice] -= 1
            if targets[choice] == 0:
                del targets[choice]
        w.commit()
        
        s = ix.searcher()
        fs = searching.Facets(s)
        for name in domain:
            fs.add_facet(name, query.Term("chapter", name))
        
        r = s.search(query.Every("chapter"), limit=None)
        self.assertEqual(fs.counts(r), dict(alfa=5, bravo=10, charlie=7, delta=2, echo=9))
        self.assertEqual(fs.categorize(r), dict(cats))
        



if __name__ == '__main__':
    unittest.main()




        
    


import unittest

from whoosh import fields, index
from whoosh.filedb.filestore import RamStorage
from whoosh.qparser import QueryParser
from whoosh.query import *


class TestQueries(unittest.TestCase):
    def test_all_terms(self):
        q = QueryParser("a").parse(u'hello b:there c:"my friend"')
        ts = set()
        q.all_terms(ts, phrases=False)
        self.assertEqual(sorted(ts), [("a", "hello"), ("b", "there")])
        ts = set()
        q.all_terms(ts, phrases=True)
        self.assertEqual(sorted(ts), [("a", "hello"), ("b", "there"),
                                      ("c", "friend"), ("c", "my")])
    
    def test_existing_terms(self):
        s = fields.Schema(key=fields.ID, value=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(key=u"a", value=u"alfa bravo charlie delta echo")
        w.add_document(key=u"b", value=u"foxtrot golf hotel india juliet")
        w.commit()
        
        r = ix.reader()
        q = QueryParser("value").parse(u'alfa hotel tango "sierra bravo"')
        
        ts = q.existing_terms(r, phrases=False)
        self.assertEqual(sorted(ts), [("value", "alfa"), ("value", "hotel")])
        
        ts = q.existing_terms(r)
        self.assertEqual(sorted(ts), [("value", "alfa"), ("value", "bravo"), ("value", "hotel")])
        
        ts = set()
        q.existing_terms(r, ts, reverse=True)
        self.assertEqual(sorted(ts), [("value", "sierra"), ("value", "tango")])
    
    def test_replace(self):
        q = And([Or([Term("a", "b"), Term("b", "c")], boost=1.2), Variations("a", "b", boost=2.0)])
        q = q.replace("b", "BB")
        self.assertEqual(q, And([Or([Term("a", "BB"), Term("b", "c")], boost=1.2), Variations("a", "BB", boost=2.0)]))
    
    def test_visitor(self):
        def visitor(q):
            if isinstance(q, (Term, Variations, FuzzyTerm)):
                q.text = q.text.upper()
            return q
        
        before = And([Not(Term("a", u"b")), Variations("a", u"c"), FuzzyTerm("a", u"d")])
        after = before.accept(visitor)
        self.assertEqual(after, And([Not(Term("a", u"B")), Variations("a", u"C"), FuzzyTerm("a", u"D")]))

    def test_simplify(self):
        s = fields.Schema(k=fields.ID, v=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(k=u"1", v=u"aardvark apple allan alfa bear bee")
        w.add_document(k=u"2", v=u"brie glue geewhiz goop julia")
        w.commit()
        
        r = ix.reader()
        q1 = And([Prefix("v", "b", boost=2.0), Term("v", "juliet")])
        q2 = And([Or([Term('v', u'bear', boost=2.0), Term('v', u'bee', boost=2.0),
                      Term('v', u'brie', boost=2.0)]), Term('v', 'juliet')])
        self.assertEqual(q1.simplify(r), q2)


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime, timedelta

from whoosh import fields, qparser, query
from whoosh.filedb.filestore import RamStorage
from whoosh.support import numeric


class TestSchema(unittest.TestCase):
    def test_schema_eq(self):
        a = fields.Schema()
        b = fields.Schema()
        self.assertEqual(a, b)

        a = fields.Schema(id=fields.ID)
        b = a.copy()
        self.assertEqual(a["id"], b["id"])
        self.assertEqual(a, b)

        c = fields.Schema(id=fields.TEXT)
        self.assertNotEqual(a, c)
    
    def test_creation1(self):
        s = fields.Schema()
        s.add("content", fields.TEXT(phrase = True))
        s.add("title", fields.TEXT(stored = True))
        s.add("path", fields.ID(stored = True))
        s.add("tags", fields.KEYWORD(stored = True))
        s.add("quick", fields.NGRAM)
        s.add("note", fields.STORED)
        
        self.assertEqual(s.names(), ["content", "note", "path", "quick", "tags", "title"])
        self.assert_("content" in s)
        self.assertFalse("buzz" in s)
        self.assert_(isinstance(s["tags"], fields.KEYWORD))
        
    def test_creation2(self):
        s = fields.Schema(a=fields.ID(stored=True),
                          b=fields.ID,
                          c=fields.KEYWORD(scorable=True))
        
        self.assertEqual(s.names(), ["a", "b", "c"])
        self.assertTrue("a" in s)
        self.assertTrue("b" in s)
        self.assertTrue("c" in s)
        
    def test_badnames(self):
        s = fields.Schema()
        self.assertRaises(fields.FieldConfigurationError, s.add, "_test", fields.ID)
        self.assertRaises(fields.FieldConfigurationError, s.add, "a f", fields.ID)
    
    def test_numeric_support(self):
        intf = fields.NUMERIC(int, shift_step=0)
        longf = fields.NUMERIC(long, shift_step=0)
        floatf = fields.NUMERIC(float, shift_step=0)
        
        def roundtrip(obj, num):
            self.assertEqual(obj.from_text(obj.to_text(num)), num)
        
        roundtrip(intf, 0)
        roundtrip(intf, 12345)
        roundtrip(intf, -12345)
        roundtrip(longf, 0)
        roundtrip(longf, 85020450482)
        roundtrip(longf, -85020450482)
        roundtrip(floatf, 0)
        roundtrip(floatf, 582.592)
        roundtrip(floatf, -582.592)
        roundtrip(floatf, -99.42)
        
        from random import shuffle
        def roundtrip_sort(obj, start, end, step):
            count = start
            rng = []
            while count < end:
                rng.append(count)
                count += step
            
            scrabled = rng[:]
            shuffle(scrabled)
            round = [obj.from_text(t) for t
                     in sorted([obj.to_text(n) for n in scrabled])]
            self.assertEqual(round, rng)
        
        roundtrip_sort(intf, -100, 100, 1)
        roundtrip_sort(longf, -58902, 58249, 43)
        roundtrip_sort(floatf, -99.42, 99.83, 2.38)
    
    def test_numeric(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               integer=fields.NUMERIC(int),
                               floating=fields.NUMERIC(float))
        ix = RamStorage().create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"a", integer=5820, floating=1.2)
        w.add_document(id=u"b", integer=22, floating=2.3)
        w.add_document(id=u"c", integer=78, floating=3.4)
        w.add_document(id=u"d", integer=13, floating=4.5)
        w.add_document(id=u"e", integer=9, floating=5.6)
        w.commit()
        
        s = ix.searcher()
        qp = qparser.QueryParser("integer", schema=schema)
        
        r = s.search(qp.parse("5820"))
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0]["id"], "a")
        
        s = ix.searcher()
        r = s.search(qp.parse("floating:4.5"))
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0]["id"], "d")
        
        q = qp.parse("integer:*")
        self.assertEqual(q.__class__, query.Every)
        self.assertEqual(q.fieldname, "integer")
        
        q = qp.parse("integer:5?6")
        self.assertEqual(q, query.NullQuery)
        
    def test_decimal_numeric(self):
        from decimal import Decimal
        
        f = fields.NUMERIC(int, decimal_places=4)
        schema = fields.Schema(id=fields.ID(stored=True), deci=f)
        ix = RamStorage().create_index(schema)
        
        self.assertEqual(f.from_text(f.to_text(Decimal("123.56"))),
                         Decimal("123.56"))
        
        w = ix.writer()
        w.add_document(id=u"a", deci=Decimal("123.56"))
        w.add_document(id=u"b", deci=Decimal("0.536255"))
        w.add_document(id=u"c", deci=Decimal("2.5255"))
        w.add_document(id=u"d", deci=Decimal("58"))
        w.commit()
        
        s = ix.searcher()
        qp = qparser.QueryParser("deci", schema=schema)
        
        r = s.search(qp.parse("123.56"))
        self.assertEqual(r[0]["id"], "a")
        
        r = s.search(qp.parse("0.536255"))
        self.assertEqual(r[0]["id"], "b")
    
    def test_numeric_parsing(self):
        schema = fields.Schema(id=fields.ID(stored=True), number=fields.NUMERIC)
        
        qp = qparser.QueryParser("number", schema=schema)
        q = qp.parse("[10 to *]")
        self.assertEqual(q, query.NullQuery)
        
        q = qp.parse("[to 400]")
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, None)
        self.assertEqual(q.end, 400)
        
        q = qp.parse("[10 to]")
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, 10)
        self.assertEqual(q.end, None)
        
        q = qp.parse("[10 to 400]")
        self.assertEqual(q.__class__, query.NumericRange)
        self.assertEqual(q.start, 10)
        self.assertEqual(q.end, 400)
        
    def test_numeric_ranges(self):
        schema = fields.Schema(id=fields.STORED, num=fields.NUMERIC)
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        
        for i in xrange(400):
            w.add_document(id=i, num=i)
        w.commit()
        
        s = ix.searcher()
        qp = qparser.QueryParser("num", schema=schema)
        
        def check(qs, target):
            q = qp.parse(qs)
            result = [s.stored_fields(d)["id"] for d in q.docs(s)]
            self.assertEqual(result, target)
        
        # Note that range() is always inclusive-exclusive
        check("[10 to 390]", range(10, 390+1))
        check("[100 to]", range(100, 400))
        check("[to 350]", range(0, 350+1))
        check("[16 to 255]", range(16, 255+1))
        check("{10 to 390]", range(11, 390+1))
        check("[10 to 390}", range(10, 390))
        check("{10 to 390}", range(11, 390))
        check("{16 to 255}", range(17, 255))
        
    def test_decimal_ranges(self):
        from decimal import Decimal
        
        schema = fields.Schema(id=fields.STORED, num=fields.NUMERIC(int, decimal_places=2))
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        count = Decimal("0.0")
        inc = Decimal("0.2")
        for _ in xrange(500):
            w.add_document(id=str(count), num=count)
            count += inc
        w.commit()
        
        s = ix.searcher()
        qp = qparser.QueryParser("num", schema=schema)
        
        def check(qs, start, end):
            q = qp.parse(qs)
            result = [s.stored_fields(d)["id"] for d in q.docs(s)]
            
            target = []
            count = Decimal(start)
            limit = Decimal(end)
            while count <= limit:
                target.append(str(count))
                count += inc
            
            self.assertEqual(result, target)
        
        check("[10.2 to 80.8]", "10.2", "80.8")
        check("{10.2 to 80.8]", "10.4", "80.8")
        check("[10.2 to 80.8}", "10.2", "80.6")
        check("{10.2 to 80.8}", "10.4", "80.6")
    
    def test_nontext_document(self):
        schema = fields.Schema(id=fields.STORED, num=fields.NUMERIC,
                               date=fields.DATETIME, even=fields.BOOLEAN)
        ix = RamStorage().create_index(schema)
        
        dt = datetime.now()
        w = ix.writer()
        for i in xrange(100):
            w.add_document(id=i, num=i, date=dt + timedelta(days=i), even=not(i % 2))
        w.commit()
        
        s = ix.searcher()
        def check(kwargs, target):
            result = [d['id'] for d in s.documents(**kwargs)]
            self.assertEqual(result, target)
        
        check({"num": 99}, [99])
        check({"date": dt + timedelta(days=30)}, [30])
        check({"even": True}, range(0, 100, 2))
    
    def test_datetime(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               date=fields.DATETIME(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        for month in xrange(1, 12):
            for day in xrange(1, 28):
                w.add_document(id=u"%s-%s" % (month, day),
                               date=datetime(2010, month, day, 14, 00, 00))
        w.commit()
        
        s = ix.searcher()
        qp = qparser.QueryParser("id", schema=schema)
        
        r = s.search(qp.parse("date:20100523"))
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0]["id"], "5-23")
        self.assertEqual(r[0]["date"].__class__, datetime)
        self.assertEqual(r[0]["date"].month, 5)
        self.assertEqual(r[0]["date"].day, 23)
        
        r = s.search(qp.parse("date:'2010 02'"))
        self.assertEqual(len(r), 27)
        
        q = qp.parse(u"date:[2010-05 TO 2010-08]")
        self.assertEqual(q.__class__, query.TermRange)
        self.assertEqual(q.start, u"201005")
        self.assertEqual(q.end, u"201008")
    
    def test_boolean(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               done=fields.BOOLEAN(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"a", done=True)
        w.add_document(id=u"b", done=False)
        w.add_document(id=u"c", done=True)
        w.add_document(id=u"d", done=False)
        w.add_document(id=u"e", done=True)
        w.commit()
        
        s = ix.searcher()
        qp = qparser.QueryParser("id", schema=schema)
        
        def all_false(ls):
            for item in ls:
                if item: return False
            return True
        
        r = s.search(qp.parse("done:true"))
        self.assertEqual(sorted([d["id"] for d in r]), ["a", "c", "e"])
        self.assertTrue(all(d["done"] for d in r))
        
        r = s.search(qp.parse("done:yes"))
        self.assertEqual(sorted([d["id"] for d in r]), ["a", "c", "e"])
        self.assertTrue(all(d["done"] for d in r))
        
        r = s.search(qp.parse("done:false"))
        self.assertEqual(sorted([d["id"] for d in r]), ["b", "d"])
        self.assertTrue(all_false(d["done"] for d in r))
        
        r = s.search(qp.parse("done:no"))
        self.assertEqual(sorted([d["id"] for d in r]), ["b", "d"])
        self.assertTrue(all_false(d["done"] for d in r))






if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

""" Contains functions and classes related to fields.
"""

import datetime, re, struct
from decimal import Decimal

from whoosh.analysis import (IDAnalyzer, RegexAnalyzer, KeywordAnalyzer,
                             StandardAnalyzer, NgramAnalyzer, Tokenizer,
                             NgramWordAnalyzer, Analyzer)
from whoosh.formats import Format, Existence, Frequency, Positions
from whoosh.support.numeric import (int_to_text, text_to_int, long_to_text,
                                    text_to_long, float_to_text, text_to_float,
                                    )


# Exceptions

class FieldConfigurationError(Exception):
    pass
class UnknownFieldError(Exception):
    pass


# Field Types

class FieldType(object):
    """Represents a field configuration.
    
    The FieldType object supports the following attributes:
    
    * format (fields.Format): the storage format for the field's contents.
    
    * vector (fields.Format): the storage format for the field's vectors
      (forward index), or None if the field should not store vectors.
    
    * scorable (boolean): whether searches against this field may be scored.
      This controls whether the index stores per-document field lengths for
      this field.
          
    * stored (boolean): whether the content of this field is stored for each
      document. For example, in addition to indexing the title of a document,
      you usually want to store the title so it can be presented as part of
      the search results.
         
    * unique (boolean): whether this field's value is unique to each document.
      For example, 'path' or 'ID'. IndexWriter.update_document() will use
      fields marked as 'unique' to find the previous version of a document
      being updated.
      
    The constructor for the base field type simply lets you supply your own
    configured field format, vector format, and scorable and stored values.
    Subclasses may configure some or all of this for you.
    
    """
    
    format = vector = scorable = stored = unique = None
    indexed = True
    __inittypes__ = dict(format=Format, vector=Format,
                         scorable=bool, stored=bool, unique=bool)
    
    def __init__(self, format, vector=None, scorable=False, stored=False,
                 unique=False):
        self.format = format
        self.vector = vector
        self.scorable = scorable
        self.stored = stored
        self.unique = unique
    
    def __repr__(self):
        temp = "%s(format=%r, vector=%r, scorable=%s, stored=%s, unique=%s)"
        return temp % (self.__class__.__name__, self.format, self.vector,
                       self.scorable, self.stored, self.unique)
    
    def __eq__(self, other):
        return all((isinstance(other, FieldType),
                    (self.format == other.format),
                    (self.vector == other.vector),
                    (self.scorable == other.scorable),
                    (self.stored == other.stored),
                    (self.unique == other.unique)))
    
    def clean(self):
        """Clears any cached information in the field and any child objects.
        """
        
        if self.format and hasattr(self.format, "clean"):
            self.format.clean()
        if self.vector and hasattr(self.vector, "clean"):
            self.vector.clean()
    
    def to_text(self, value):
        """Returns a textual representation of the value. Non-textual fields
        (such as NUMERIC and DATETIME) will override this to encode objects
        as text.
        """
        
        return value
    
    def index(self, value, **kwargs):
        """Returns an iterator of (termtext, frequency, weight, encoded_value)
        tuples.
        """
        
        if not self.format:
            raise Exception("%s field cannot index without a format" % self.__class__)
        if not isinstance(value, unicode):
            raise ValueError("%r is not unicode" % value)
        return self.format.word_values(value, mode="index", **kwargs)
    
    def process_text(self, qstring, mode='', **kwargs):
        """Returns an iterator of token strings corresponding to the given
        string.
        """
        
        if not self.format:
            raise Exception("%s field has no format" % self)
        return (t.text for t
                in self.format.analyze(qstring, mode=mode, **kwargs))
        
    def self_parsing(self):
        """Subclasses should override this method to return True if they want
        the query parser to call the field's ``parse_query()`` method instead
        of running the analyzer on text in this field. This is useful where
        the field needs full control over how queries are interpreted, such
        as in the numeric field type.
        """
        
        return False
    
    def parse_query(self, fieldname, qstring, boost=1.0):
        """When ``self_parsing()`` returns True, the query parser will call
        this method to parse basic query text.
        """
        
        raise NotImplementedError(self.__class__.__name__)
    
    def parse_range(self, fieldname, start, end, startexcl, endexcl, boost=1.0):
        """When ``self_parsing()`` returns True, the query parser will call
        this method to parse range query text. If this method returns None
        instead of a query object, the parser will fall back to parsing the
        start and end terms using process_text().
        """
        
        return None
    

class ID(FieldType):
    """Configured field type that indexes the entire value of the field as one
    token. This is useful for data you don't want to tokenize, such as the path
    of a file.
    """
    
    __inittypes__ = dict(stored=bool, unique=bool, field_boost=float)
    
    def __init__(self, stored=False, unique=False, field_boost=1.0):
        """
        :param stored: Whether the value of this field is stored with the document.
        """
        self.format = Existence(analyzer=IDAnalyzer(), field_boost=field_boost)
        self.stored = stored
        self.unique = unique


class IDLIST(FieldType):
    """Configured field type for fields containing IDs separated by whitespace
    and/or puntuation.
    """
    
    __inittypes__ = dict(stored=bool, unique=bool, expression=bool, field_boost=float)
    
    def __init__(self, stored=False, unique=False, expression=None, field_boost=1.0):
        """
        :param stored: Whether the value of this field is stored with the
            document.
        :param unique: Whether the value of this field is unique per-document.
        :param expression: The regular expression object to use to extract
            tokens. The default expression breaks tokens on CRs, LFs, tabs,
            spaces, commas, and semicolons.
        """
        
        expression = expression or re.compile(r"[^\r\n\t ,;]+")
        analyzer = RegexAnalyzer(expression=expression)
        self.format = Existence(analyzer=analyzer, field_boost=field_boost)
        self.stored = stored
        self.unique = unique


class NUMERIC(FieldType):
    """Special field type that lets you index int, long, or floating point
    numbers in relatively short fixed-width terms. The field converts numbers
    to sortable text for you before indexing.
    
    You specify the numeric type of the field when you create the NUMERIC
    object. The default is ``int``.
    
    >>> schema = Schema(path=STORED, position=NUMERIC(long))
    >>> ix = storage.create_index(schema)
    >>> w = ix.writer()
    >>> w.add_document(path="/a", position=5820402204)
    >>> w.commit()
    
    You can also use the NUMERIC field to store Decimal instances by specifying
    a type of ``int`` or ``long`` and the ``decimal_places`` keyword argument.
    This simply multiplies each number by ``(10 ** decimal_places)`` before
    storing it as an integer. Of course this may throw away decimal prcesision
    (by truncating, not rounding) and imposes the same maximum value limits as
    ``int``/``long``, but these may be acceptable for certain applications.
    
    >>> from decimal import Decimal
    >>> schema = Schema(path=STORED, position=NUMERIC(int, decimal_places=4))
    >>> ix = storage.create_index(schema)
    >>> w = ix.writer()
    >>> w.add_document(path="/a", position=Decimal("123.45")
    >>> w.commit()
    """
    
    def __init__(self, type=int, stored=False, unique=False, field_boost=1.0,
                 decimal_places=0, shift_step=4, signed=True):
        """
        :param type: the type of numbers that can be stored in this field: one
            of ``int``, ``long``, ``float``, or ``Decimal``.
        :param stored: Whether the value of this field is stored with the
            document.
        :param unique: Whether the value of this field is unique per-document.
        :param decimal_places: specifies the number of decimal places to save
            when storing Decimal instances as ``int`` or ``float``.
        :param shift_steps: The number of bits of precision to shift away at
            each tiered indexing level. Values should generally be 1-8. Lower
            values yield faster searches but take up more space. A value
            of `0` means no tiered indexing.
        :param signed: Whether the numbers stored in this field may be
            negative.
        """
        
        self.type = type
        if self.type is int:
            self._to_text = int_to_text
            self._from_text = text_to_int
        elif self.type is long:
            self._to_text = long_to_text
            self._from_text = text_to_long
        elif self.type is float:
            self._to_text =  float_to_text
            self._from_text = text_to_float
        elif self.type is Decimal:
            raise TypeError("To store Decimal instances, set type to int or "
                            "float and use the decimal_places argument")
        else:
            raise TypeError("%s field type can't store %r" % (self.__class__,
                                                              self.type))
        
        self.stored = stored
        self.unique = unique
        self.decimal_places = decimal_places
        self.shift_step = shift_step
        self.signed = signed
        self.format = Existence(analyzer=IDAnalyzer(), field_boost=field_boost)
    
    def _tiers(self, num):
        t = self.type
        if t is int:
            bitlen = 32
        else:
            bitlen = 64
        
        for shift in xrange(0, bitlen, self.shift_step):
            yield self.to_text(num, shift=shift)
    
    def index(self, num):
        # word, freq, weight, valuestring
        if self.shift_step:
            return [(txt, 1, 1.0, '') for txt in self._tiers(num)]
        else:
            return [(self.to_text(num), 1, 1.0, '')]
    
    def prepare_number(self, x):
        if x is None: return x
        if self.decimal_places:
            x = Decimal(x)
            x *= 10 ** self.decimal_places
        x = self.type(x)
        return x
    
    def unprepare_number(self, x):
        if self.decimal_places:
            s = str(x)
            x = Decimal(s[:-self.decimal_places] + "." + s[-self.decimal_places:])
        return x
    
    def to_text(self, x, shift=0):
        return self._to_text(self.prepare_number(x), shift=shift)
    
    def from_text(self, t):
        x = self._from_text(t)
        return self.unprepare_number(x)
    
    def process_text(self, text, **kwargs):
        return (self.to_text(text),)
    
    def self_parsing(self):
        return True
    
    def parse_query(self, fieldname, qstring, boost=1.0):
        from whoosh import query
        from whoosh.qparser import QueryParserError
        
        if qstring == "*":
            return query.Every(fieldname, boost=boost)
        
        try:
            text = self.to_text(qstring)
        except Exception, e:
            raise QueryParserError(e)
        
        return query.Term(fieldname, text, boost=boost)
    
    def parse_range(self, fieldname, start, end, startexcl, endexcl, boost=1.0):
        from whoosh import query
        from whoosh.qparser import QueryParserError
        
        try:
            if start is not None:
                start = self.from_text(self.to_text(start))
            if end is not None:
                end = self.from_text(self.to_text(end))
        except Exception, e:
            raise QueryParserError(e)
        
        return query.NumericRange(fieldname, start, end, startexcl, endexcl,
                                  boost=boost)
    

class DATETIME(FieldType):
    """Special field type that lets you index datetime objects. The field
    converts the datetime objects to sortable text for you before indexing.
    
    >>> schema = Schema(path=STORED, date=DATETIME)
    >>> ix = storage.create_index(schema)
    >>> w = ix.writer()
    >>> w.add_document(path="/a", date=datetime.now())
    >>> w.commit()
    """
    
    __inittypes__ = dict(stored=bool, unique=bool)
    
    def __init__(self, stored=False, unique=False):
        """
        :param stored: Whether the value of this field is stored with the
            document.
        :param unique: Whether the value of this field is unique per-document.
        """
        
        self.stored = stored
        self.unique = unique
        self.format = Existence(None)
    
    def to_text(self, dt):
        if not isinstance(dt, datetime.datetime):
            raise ValueError("%r is not a datetime object" % dt)
        text = dt.isoformat() # 2010-02-02T17:06:19.109000
        text = text.replace(" ", "").replace(":", "").replace("-", "").replace(".", "")
        return text
    
    def index(self, dt):
        # word, freq, weight, valuestring
        return [(self.to_text(dt), 1, 1.0, '')]
    
    def process_text(self, text, **kwargs):
        text = text.replace(" ", "").replace(":", "").replace("-", "").replace(".", "")
        return (text, )
    
    def self_parsing(self):
        return True
    
    def parse_query(self, fieldname, qstring, boost=1.0):
        text = self.process_text(qstring)[0]
        from whoosh import query
        return query.Prefix(fieldname, text, boost=boost)
    

class BOOLEAN(FieldType):
    """Special field type that lets you index boolean values (True and False).
    The field converts the boolean values to text for you before indexing.
    
    >>> schema = Schema(path=STORED, done=BOOLEAN)
    >>> ix = storage.create_index(schema)
    >>> w = ix.writer()
    >>> w.add_document(path="/a", done=False)
    >>> w.commit()
    """
    
    strings = (u"t", u"f")
    trues = frozenset((u"t", u"true", u"yes", u"1"))
    falses = frozenset((u"f", u"false", u"no", u"0"))
    
    __inittypes__ = dict(stored=bool)
    
    def __init__(self, stored=False):
        """
        :param stored: Whether the value of this field is stored with the
            document.
        """
        
        self.stored = stored
        self.format = Existence(None)
    
    def to_text(self, bit):
        if isinstance(bit, basestring):
            if bit in self.trues:
                bit = True
            else:
                bit = False
        elif not isinstance(bit, bool):
            raise ValueError("%r is not a boolean")
        return self.strings[int(bit)]
    
    def index(self, bit):
        bit = bool(bit)
        # word, freq, weight, valuestring
        return [(self.strings[int(bit)], 1, 1.0, '')]
    
    def self_parsing(self):
        return True
    
    def parse_query(self, fieldname, qstring, boost=1.0):
        from whoosh import query
        text = None
        
        if qstring == "*":
            return query.Every(fieldname, boost=boost)
        
        if qstring in self.falses:
            text = self.strings[0]
        elif qstring in self.trues:
            text = self.strings[1]
        
        if text is None:
            return query.NullQuery
        else:
            return query.Term(fieldname, text, boost=boost)
    

class STORED(FieldType):
    """Configured field type for fields you want to store but not index.
    """
    
    indexed = False
    stored = True
    
    def __init__(self):
        pass
    

class KEYWORD(FieldType):
    """Configured field type for fields containing space-separated or
    comma-separated keyword-like data (such as tags). The default is to not
    store positional information (so phrase searching is not allowed in this
    field) and to not make the field scorable.
    """
    
    __inittypes__ = dict(stored=bool, lowercase=bool, commas=bool, scorable=bool,
                         unique=bool, field_boost=float)
    
    def __init__(self, stored=False, lowercase=False, commas=False,
                 scorable=False, unique=False, field_boost=1.0):
        """
        :param stored: Whether to store the value of the field with the
            document.
        :param comma: Whether this is a comma-separated field. If this is False
            (the default), it is treated as a space-separated field.
        :param scorable: Whether this field is scorable.
        """
        
        ana = KeywordAnalyzer(lowercase=lowercase, commas=commas)
        self.format = Frequency(analyzer=ana, field_boost=field_boost)
        self.scorable = scorable
        self.stored = stored
        self.unique = unique


class TEXT(FieldType):
    """Configured field type for text fields (for example, the body text of an
    article). The default is to store positional information to allow phrase
    searching. This field type is always scorable.
    """
    
    __inittypes__ = dict(analyzer=Analyzer, phrase=bool, vector=Format,
                         stored=bool, field_boost=float)
    
    def __init__(self, analyzer=None, phrase=True, vector=None, stored=False,
                 field_boost=1.0):
        """
        :param analyzer: The analysis.Analyzer to use to index the field
            contents. See the analysis module for more information. If you omit
            this argument, the field uses analysis.StandardAnalyzer.
        :param phrase: Whether the store positional information to allow phrase
            searching.
        :param vector: A :class:`whoosh.formats.Format` object to use to store
            term vectors. By default, fields do not store term vectors.
        :param stored: Whether to store the value of this field with the
            document. Since this field type generally contains a lot of text,
            you should avoid storing it with the document unless you need to,
            for example to allow fast excerpts in the search results.
        """
        
        ana = analyzer or StandardAnalyzer()
        
        if phrase:
            formatclass = Positions
        else:
            formatclass = Frequency
        self.format = formatclass(analyzer=ana, field_boost=field_boost)
        self.vector = vector
        
        self.scorable = True
        self.stored = stored


class NGRAM(FieldType):
    """Configured field that indexes text as N-grams. For example, with a field
    type NGRAM(3,4), the value "hello" will be indexed as tokens
    "hel", "hell", "ell", "ello", "llo". This field chops the entire 
    """
    
    __inittypes__ = dict(minsize=int, maxsize=int, stored=bool, field_boost=float)
    scorable = True
    
    def __init__(self, minsize=2, maxsize=4, stored=False, field_boost=1.0):
        """
        :param minsize: The minimum length of the N-grams.
        :param maxsize: The maximum length of the N-grams.
        :param stored: Whether to store the value of this field with the
            document. Since this field type generally contains a lot of text,
            you should avoid storing it with the document unless you need to,
            for example to allow fast excerpts in the search results.
        """
        
        self.format = Frequency(analyzer=NgramAnalyzer(minsize, maxsize),
                                field_boost=field_boost)
        self.stored = stored


class NGRAMWORDS(FieldType):
    """Configured field that breaks text into words, lowercases, and then chops
    the words into N-grams.
    """
    
    __inittypes__ = dict(minsize=int, maxsize=int, stored=bool,
                         field_boost=float, tokenizer=Tokenizer)
    scorable = True
    
    def __init__(self, minsize=2, maxsize=4, stored=False, field_boost=1.0,
                 tokenizer=None, at=None):
        """
        :param minsize: The minimum length of the N-grams.
        :param maxsize: The maximum length of the N-grams.
        :param stored: Whether to store the value of this field with the
            document. Since this field type generally contains a lot of text,
            you should avoid storing it with the document unless you need to,
            for example to allow fast excerpts in the search results.
        :param tokenizer: an instance of :class:`whoosh.analysis.Tokenizer`
            used to break the text into words.
        """
        
        analyzer = NgramWordAnalyzer(minsize, maxsize, tokenizer, at=at)
        self.format = Frequency(analyzer=analyzer, field_boost=field_boost)
        self.stored = stored


# Schema class

class Schema(object):
    """Represents the collection of fields in an index. Maps field names to
    FieldType objects which define the behavior of each field.
    
    Low-level parts of the index use field numbers instead of field names for
    compactness. This class has several methods for converting between the
    field name, field number, and field object itself.
    """
    
    def __init__(self, **fields):
        """ All keyword arguments to the constructor are treated as fieldname =
        fieldtype pairs. The fieldtype can be an instantiated FieldType object,
        or a FieldType sub-class (in which case the Schema will instantiate it
        with the default constructor before adding it).
        
        For example::
        
            s = Schema(content = TEXT,
                       title = TEXT(stored = True),
                       tags = KEYWORD(stored = True))
        """
        
        self._fields = {}
        
        for name in sorted(fields.keys()):
            self.add(name, fields[name])
    
    def copy(self):
        """Returns a shallow copy of the schema. The field instances are not
        deep copied, so they are shared between schema copies.
        """
        
        s = self.__class__()
        s._fields = self._fields.copy()
        return s
    
    def __eq__(self, other):
        return (isinstance(other, Schema)
                and self._fields == other._fields)
    
    def __repr__(self):
        return "<Schema: %s>" % repr(self._fields.keys())
    
    def __iter__(self):
        """Returns the field objects in this schema.
        """
        
        return self._fields.itervalues()
    
    def __getitem__(self, name):
        """Returns the field associated with the given field name.
        """
        
        return self._fields[name]
        
    def __len__(self):
        """Returns the number of fields in this schema.
        """
        
        return len(self._fields)
    
    def __contains__(self, fieldname):
        """Returns True if a field by the given name is in this schema.
        """
        
        return fieldname in self._fields
    
    def items(self):
        """Returns a list of ("fieldname", field_object) pairs for the fields
        in this schema.
        """
        
        return sorted(self._fields.items())
        
    def names(self):
        """Returns a list of the names of the fields in this schema.
        """
        return sorted(self._fields.keys())
    
    def clean(self):
        for field in self:
            field.clean()
    
    def add(self, name, fieldtype):
        """Adds a field to this schema. This is a low-level method; use keyword
        arguments to the Schema constructor to create the fields instead.
        
        :param name: The name of the field.
        :param fieldtype: An instantiated fields.FieldType object, or a
            FieldType subclass. If you pass an instantiated object, the schema
            will use that as the field configuration for this field. If you
            pass a FieldType subclass, the schema will automatically
            instantiate it with the default constructor.
        """
        
        if name.startswith("_"):
            raise FieldConfigurationError("Field names cannot start with an underscore")
        if " " in name:
            raise FieldConfigurationError("Field names cannot contain spaces")
        elif name in self._fields:
            raise FieldConfigurationError("Schema already has a field named %s" % name)
        
        if type(fieldtype) is type:
            try:
                fieldtype = fieldtype()
            except Exception, e:
                raise FieldConfigurationError("Error: %s instantiating field %r: %r" % (e, name, fieldtype))
        
        if not isinstance(fieldtype, FieldType):
            raise FieldConfigurationError("%r is not a FieldType object" % fieldtype)
        
        self._fields[name] = fieldtype
        
    def remove(self, fieldname):
        del self._fields[fieldname]
        
    def has_vectored_fields(self):
        """Returns True if any of the fields in this schema store term vectors.
        """
        
        return any(ftype.vector for ftype in self)
    
    def has_scorable_fields(self):
        return any(ftype.scorable for ftype in self)
    
    def stored_names(self):
        """Returns a list of the names of fields that are stored.
        """
        
        return [name for name, field in self.items() if field.stored]

    def scorable_names(self):
        """Returns a list of the names of fields that store field
        lengths.
        """
        
        return [name for name, field in self.items() if field.scorable]

    def vector_names(self):
        """Returns a list of the names of fields that store vectors.
        """
        
        return [name for name, field in self.items() if field.vector]

    def analyzer(self, fieldname):
        """Returns the content analyzer for the given fieldname, or None if
        the field has no analyzer
        """
        
        field = self[fieldname]
        if field.format and field.format.analyzer:
            return field.format.analyzer
        

    
    
    
    
    
    
    




#===============================================================================
# Copyright 2008 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

__version__ = (1, 2, 4)


def versionstring(build=True, extra=True):
    """Returns the version number of Whoosh as a string.
    
    :param build: Whether to include the build number in the string.
    :param extra: Whether to include alpha/beta/rc etc. tags. Only
        checked if build is True.
    :rtype: str
    """
    
    if build:
        first = 3
    else:
        first = 2
    
    s = ".".join(str(n) for n in __version__[:first])
    if build and extra:
        s += "".join(str(n) for n in __version__[3:])
    
    return s




import gzip, os.path
from optparse import OptionParser

from whoosh import analysis, fields, index, qparser, query
from whoosh.util import now

#ana = analysis.StemmingAnalyzer()
ana = analysis.StandardAnalyzer()
schema = fields.Schema(id=fields.ID(stored=True),
                       headline=fields.STORED,
                       text=fields.TEXT(analyzer=ana, stored=True))

def do_index(file, indexname, **kwargs):
    print "Indexing..."
    if not os.path.exists(indexname):
        os.mkdir(indexname)
    ix = index.create_in(indexname, schema)
    
    t = now()
    w = ix.writer(**kwargs)
    for line in gzip.GzipFile(file, "rb"):
        id, text = line.decode("latin1").split("\t")
        w.add_document(id=id, text=text, headline=text[:70])
    print "Spool:", now() - t
    ct = now()
    w.commit()
    print "Commit:", now() - ct
    print "Total:", now() - t


def do_search(indexname, q, limit=10, showbody=False):
    ix = index.open_dir(indexname)
    s = ix.searcher()
    q = qparser.QueryParser("text", schema=s.schema).parse(q)
    print "query=", q
    r = s.search(q, limit=limit)
    print "result=", r
    for hit in r:
        print hit.pos, hit["headline"]
        if showbody:
            print hit["text"]
        
if __name__ == "__main__":
    parser = OptionParser()
    parser.add_option("-i", "--index", dest="index",
                      help="Index the documents", default=False,
                      action="store_true")
    parser.add_option("-s", "--source", dest="source", metavar="FILENAME",
                      help="File containing the corpus date.",
                      default="reuters21578.txt.gz")
    parser.add_option("-d", "--dir", dest="dir", metavar="DIRNAME",
                      help="Directory in which to store files, index, etc.",
                      default=".")
    parser.add_option("-n", "--name", dest="indexname",
                      help="Name of the index directory",
                      default="reuters_index")
    parser.add_option("-m", "--mb", dest="limitmb",
                      help="Memory size, in MB",
                      default="128")
    parser.add_option("-p", "--procs", dest="procs",
                      help="Use this many processors to index.",
                      default="1")
    parser.add_option("-l", "--limit", dest="limit",
                      help="Maximum number of results to display for a search.",
                      default="10")
    parser.add_option("-b", "--body", dest="showbody",
                      help="Show the body of emails found by a search.",
                      default=False, action="store_true")
    parser.add_option("-t", "--tempdir", dest="tempdir",
                      help="Directory to use for temp file storage",
                      default=None)
    
    options, args = parser.parse_args()
    
    if options.index:
        do_index(options.source, options.indexname, procs=int(options.procs),
                 limitmb=int(options.limitmb), dir=options.tempdir)
        
    if args:
        qs = " ".join(args).decode("utf8")
        print "Query string=", repr(qs)
        do_search(options.indexname, qs, limit=int(options.limit),
                  showbody=options.showbody)
    
    



from __future__ import division
from zlib import compress, decompress
from email import message_from_string
import os.path, tarfile
from marshal import dump, load
from optparse import OptionParser
from urllib import urlretrieve

from whoosh import analysis, fields, index, qparser
from whoosh.util import now


# http://www.cs.cmu.edu/~enron/
enron_archive_url = "http://www.cs.cmu.edu/~enron/enron_mail_082109.tar.gz"
enron_archive_filename = "enron_mail_082109.tar.gz"
cache_filename = "enron_cache.pickle"

ana = analysis.StemmingAnalyzer(maxsize=40)
schema = fields.Schema(body=fields.TEXT(analyzer=ana, stored=True),
                       date=fields.ID(stored=True),
                       frm=fields.ID(stored=True),
                       to=fields.IDLIST(stored=True),
                       subject=fields.TEXT(stored=True),
                       cc=fields.IDLIST,
                       bcc=fields.IDLIST)

header_to_field = {"Date": "date", "From": "frm", "To": "to",
                   "Subject": "subject", "Cc": "cc", "Bcc": "bcc"}


# Functions for downloading and then reading the email archive and caching
# the messages in an easier-to-digest format

def download_archive(archive):
    print "Downloading Enron email archive to %r..." % archive
    t = now()
    urlretrieve(enron_archive_url, archive)
    print "Downloaded in ", now() - t, "seconds"
    
def get_texts(archive):
    archive = tarfile.open(archive, "r:gz")
    while True:
        entry = archive.next()
        archive.members = []
        if entry is None:
            break
        f = archive.extractfile(entry)
        if f is not None:
            text = f.read()
            yield text

def get_messages(archive, headers=True):
    for text in get_texts(archive):
        message = message_from_string(text)
        body = message.as_string().decode("latin_1")
        blank = body.find("\n\n")
        if blank > -1:
            body = body[blank+2:]
        d = {"body": body}
        if headers:
            for k in message.keys():
                fn = header_to_field.get(k)
                if not fn: continue
                v = message.get(k).strip()
                if v:
                    d[fn] = v.decode("latin_1")
        yield d
        
def cache_messages(archive, cache):
    print "Caching messages in %s..." % cache
    
    if not os.path.exists(archive):
        raise Exception("Archive file %r does not exist" % archive)
    
    t = now()
    f = open(cache, "wb")
    c = 0
    for d in get_messages(archive):
        c += 1
        dump(d, f)
        if not c % 1000: print c
    f.close()
    print "Cached messages in ", now() - t, "seconds"


# Functions for reading the cached messages

def get_cached_messages(cache):
    f = open(cache, "rb")
    try:
        while True:
            d = load(f)
            yield d
    except EOFError:
        pass
    f.close()


# Main function for indexing the cached messages

def do_index(cache, indexname, chunk=1000, skip=1, upto=600000, **kwargs):
    print "Indexing..."
    if not os.path.exists(indexname):
        os.mkdir(indexname)
    ix = index.create_in(indexname, schema)
    
    w = ix.writer(**kwargs)
    starttime = chunkstarttime = now()
    c = 0
    skipc = skip
    for d in get_cached_messages(cache):
        skipc -= 1
        if not skipc:
            d["_stored_body"] = compress(d["body"], 9)
            w.add_document(**d)
            skipc = skip
            c += 1
            if c > upto:
                break
            if not c % chunk:
                t = now()
                sofar = t - starttime
                print "Indexed %d messages, %f for %d, %f total, %f docs/s" % (c, t - chunkstarttime, chunk, sofar, c/sofar)
                schema.clean()
                chunkstarttime = t
    spooltime = now()
    print "Spool", spooltime - starttime
    w.commit()
    committime = now()
    print "Commit", (committime - spooltime)
    print "Total", (committime - starttime), "for", c


# Main function for testing the archive

def do_search(indexname, q, limit=10, showbody=False):
    ix = index.open_dir(indexname)
    s = ix.searcher()
    q = qparser.QueryParser("body", schema=s.schema).parse(q)
    print "query=", q
    r = s.search(q, limit=limit)
    print "result=", r
    t = now()
    for hit in r:
        print hit.pos, hit.get("subject")
        if showbody:
            print decompress(hit["body"])
    print "Print time:", now() - t


if __name__=="__main__":
    parser = OptionParser()
    parser.add_option("-d", "--dir", dest="dir", metavar="DIRNAME",
                      help="directory in which to store files, index, etc.",
                      default=".")
    parser.add_option("-s", "--setup", dest="setup", action="store_true",
                      help="Download and cache the document archive if necessary.",
                      default=False)
    parser.add_option("-i", "--index", dest="index", action="store_true",
                      help="Index the documents.",
                      default=False)
    parser.add_option("-n", "--name", dest="indexname",
                      help="Name of the index directory",
                      default="enron_index")
    parser.add_option("-m", "--mb", dest="limitmb",
                      help="Memory size, in MB",
                      default="128")
    parser.add_option("-c", "--chunk", dest="chunk",
                      help="Report indexing progress in chunks of this many documents.",
                      default="1000")
    parser.add_option("-k", "--skip", dest="skip",
                      help="Skip this many documents before indexing a document.",
                      default="1")
    parser.add_option("-u", "--upto", dest="upto",
                      help="Only index up to this document.",
                      default="600000")
    parser.add_option("-p", "--procs", dest="procs",
                      help="Use this many processors to index.",
                      default="1")
    parser.add_option("-l", "--limit", dest="limit",
                      help="Maximum number of results to display for a search.",
                      default="10")
    parser.add_option("-b", "--body", dest="showbody",
                      help="Show the body of emails found by a search.",
                      default=False, action="store_true")
    parser.add_option("-t", "--tempdir", dest="tempdir",
                      help="Directory to use for temp file storage",
                      default=None)
    options, args = parser.parse_args()
    
    archive = os.path.abspath(os.path.join(options.dir, enron_archive_filename))
    cache = os.path.abspath(os.path.join(options.dir, cache_filename))
    
    if options.setup:
        if not os.path.exists(archive):
            download_archive(archive)
        else:
            print "Archive is OK"
        
        if not os.path.exists(cache):
            cache_messages(archive, cache)
        else:
            print "Cache is OK"
    
    procs = int(options.procs)
    if options.index:
        do_index(cache, options.indexname, chunk=int(options.chunk),
                 skip=int(options.skip), upto=int(options.upto),
                 procs=int(options.procs), limitmb=int(options.limitmb),
                 dir=options.tempdir)
    
    if args:
        qs = " ".join(args).decode("utf8")
        print "Query string=", repr(qs)
        do_search(options.indexname, qs, limit=int(options.limit),
                  showbody=options.showbody)
    



import unittest
from datetime import datetime

from whoosh.qparser.dateparse import *


basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
english = English()


class TestDateParser(unittest.TestCase):
#    def test_regex(self):
#        e = Regex("(?P<num>[0-9]+)")
#        props = e.parse("456", basedate)
#        self.assertEqual(props.num, 456)
#    
#    def test_sequence(self):
#        s = Sequence(Regex("(?P<name>[a-z]+)"), Regex("(?P<num>[0-9])"))
#        output = s.parse("johnny5", basedate)
#        self.assertTrue(output.__class__, Props)
#        self.assertEqual(output.name, "johnny")
#        self.assertEqual(output.num, 5)
#        
#        self.assertEqual(s.parse("5times"), None)
#    
#    def test_optional(self):
#        s = "(?P<name>[a-z]+)" + Optional("(?P<num>[0-9])") + "\\."
#        self.assertNotEqual(s.parse("johnny.", basedate), None)
#        self.assertNotEqual(s.parse("johnny5.", basedate), None)
#        self.assertEqual(s.parse("johnny5", basedate), None)
#    
#    def test_choice(self):
#        c = Choice("(?P<where>here)", "(?P<where>there)")
#        p = c.parse("here")
#        self.assertEqual(p.__class__, Props)
#        self.assertEqual(p.where, "here")
#        p = c.parse("there")
#        self.assertEqual(p.__class__, Props)
#        self.assertEqual(p.where, "there")
#        p = c.parse("anywhere")
#        self.assertEqual(p, None)
    
    def assert_atime(self, at, **kwargs):
        self.assertEqual(at.__class__, atime)
        for key in atime.units:
            if key in kwargs:
                self.assertEqual(getattr(at, key), kwargs[key])
            else:
                self.assertEqual(getattr(at, key), None)
                
    def assert_timespan(self, ts, sargs, eargs):
        self.assert_atime(ts.start, **sargs)
        self.assert_atime(ts.end, **eargs)
        
    def assert_datespan(self, ts, startdate, enddate):
        self.assertEqual(ts.start, startdate)
        self.assertEqual(ts.end, enddate)
    
    def test_time(self, t=english.time):
        self.assert_atime(t.date("13:05", basedate), hour=13, minute=5)
        self.assertEqual(t.date("28:91", basedate), None)
        
        self.assert_atime(t.date("3pm", basedate), hour=15)
        self.assert_atime(t.date("3 pm", basedate), hour=15)
        self.assert_atime(t.date("3am", basedate), hour=3)
        self.assert_atime(t.date("3:15 am", basedate), hour=3, minute=15)
        self.assert_atime(t.date("12:45am", basedate), hour=0, minute=45)
        self.assert_atime(t.date("12:45pm", basedate), hour=12, minute=45)
        self.assert_atime(t.date("5:45:05 pm", basedate), hour=17, minute=45, second=5)
        
        self.assert_atime(t.date("noon", basedate), hour=12, minute=0, second=0, microsecond=0)
        self.assert_atime(t.date("midnight", basedate), hour=0, minute=0, second=0, microsecond=0)
    
    def test_date(self, d=english.date):
        self.assert_atime(d.date("25 may 2011", basedate), year=2011, month=5, day=25)
        self.assert_atime(d.date("may 2 2011", basedate), year=2011, month=5, day=2)
        self.assert_atime(d.date("2011 25 may", basedate), year=2011, month=5, day=25)
        self.assert_atime(d.date("2011 may 5", basedate), year=2011, month=5, day=5)
        
        self.assert_atime(d.date("apr", basedate), month=4)
        self.assert_atime(d.date("september", basedate), month=9)
        self.assert_atime(d.date("2001", basedate), year=2001)
        self.assert_atime(d.date("july 2525", basedate), year=2525, month=7)
        self.assert_atime(d.date("nov 30", basedate), month=11, day=30)
        self.assertEqual(d.date("25 2525", basedate), None)
        
        self.assert_atime(d.date("25 may, 2011", basedate), year=2011, month=5, day=25)
        self.assert_atime(d.date("may 2nd, 2011", basedate), year=2011, month=5, day=2)
        self.assert_atime(d.date("2011, 25 may", basedate), year=2011, month=5, day=25)
        self.assert_atime(d.date("2011, may 5th", basedate), year=2011, month=5, day=5)
        
        self.assert_atime(d.date("today", basedate), year=2010, month=9, day=20)
        self.assert_atime(d.date("tomorrow", basedate), year=2010, month=9, day=21)
        self.assert_atime(d.date("yesterday", basedate), year=2010, month=9, day=19)
        self.assert_atime(d.date("this month", basedate), year=2010, month=9)
        self.assert_atime(d.date("this year", basedate), year=2010)
        
        self.assertEqual(d.date("now", basedate), basedate)
        
    def test_plustime(self, rt=english.plustime):
        rt = english.plustime
        
        self.assertEqual(rt.date("+1hr", basedate),
                         basedate + timedelta(hours=1))
        self.assertEqual(rt.date("+5mins", basedate),
                         basedate + timedelta(minutes=5))
        self.assertEqual(rt.date("+20s", basedate),
                         basedate + timedelta(seconds=20))
        
        self.assertEqual(rt.date("- 2 h", basedate),
                         basedate + timedelta(hours=-2))
        self.assertEqual(rt.date("- 25 minutes", basedate),
                         basedate + timedelta(minutes=-25))
        self.assertEqual(rt.date("-400 secs", basedate),
                         basedate + timedelta(seconds=-400))
        
        self.assertEqual(rt.date("+1hr 5m", basedate),
                         basedate + timedelta(hours=1, minutes=5))
        self.assertEqual(rt.date("-8hr 12m", basedate),
                         basedate + timedelta(hours=-8, minutes=-12))
        self.assertEqual(rt.date("+1hr 5s", basedate),
                         basedate + timedelta(hours=1, seconds=5))
        self.assertEqual(rt.date("+1hr 12m 5s", basedate),
                         basedate + timedelta(hours=1, minutes=12, seconds=5))
        self.assertEqual(rt.date("-1hr 5s", basedate),
                         basedate + timedelta(hours=-1, seconds=-5))
        self.assertEqual(rt.date("-1hr 12m 5s", basedate),
                         basedate + timedelta(hours=-1, minutes=-12, seconds=-5))
        
    def test_relative_days(self):
        # "next monday" on monday
        self.assertEqual(relative_days(0, 0, 1), 7)
        # "last monday" on monday
        self.assertEqual(relative_days(0, 0, -1), -7)
        # "next tuesday" on wednesday
        self.assertEqual(relative_days(2, 1, 1), 6)
        # "last tuesday" on wednesay
        self.assertEqual(relative_days(2, 1, -1), -1)
        # "last monday" on sunday
        self.assertEqual(relative_days(6, 0, -1), -6)
        # "next monday" on sunday
        self.assertEqual(relative_days(6, 0, 1), 1)
        # "next wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, 1), 1)
        # "last wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, -1), -6)
        # "last wednesday" on thursday
        self.assertEqual(relative_days(3, 2, -1), -1)
        # "next wednesday" on thursday
        self.assertEqual(relative_days(3, 2, 1), 6)
        # "last wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, -1), -6)
        # "next wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, 1), 1)
        
    def test_dayname(self, p=english.dayname):
        self.assert_atime(p.date("next tuesday", basedate), year=2010, month=9, day=21)
        self.assert_atime(p.date("last tuesday", basedate), year=2010, month=9, day=14)
        self.assert_atime(p.date("next sunday", basedate), year=2010, month=9, day=26)
        self.assert_atime(p.date("last sun", basedate), year=2010, month=9, day=19)
        self.assert_atime(p.date("next th", basedate), year=2010, month=9, day=23)
        
    def test_reldate(self, p=english.plusdate):
        self.assertEqual(p.date("+1y", basedate),
                         basedate + relativedelta(years=1))
        self.assertEqual(p.date("+2mo", basedate),
                         basedate + relativedelta(months=2))
        self.assertEqual(p.date("+3w", basedate),
                         basedate + relativedelta(weeks=3))
        self.assertEqual(p.date("+5d", basedate),
                         basedate + relativedelta(days=5))
        self.assertEqual(p.date("+5days", basedate),
                         basedate + relativedelta(days=5))
        
        self.assertEqual(p.date("-6yr", basedate),
                         basedate + relativedelta(years=-6))
        self.assertEqual(p.date("- 7 mons", basedate),
                         basedate + relativedelta(months=-7))
        self.assertEqual(p.date("-8 wks", basedate),
                         basedate + relativedelta(weeks=-8))
        self.assertEqual(p.date("- 9 dy", basedate),
                         basedate + relativedelta(days=-9))
        
        
        self.assertEqual(p.date("+1y 12mo 400d", basedate),
                         basedate + relativedelta(years=1, months=12, days=400))
        self.assertEqual(p.date("-7mo 8d", basedate),
                         basedate + relativedelta(months=-7, days=-8))
        self.assertEqual(p.date("+5wks 2d", basedate),
                         basedate + relativedelta(weeks=5, days=2))
        self.assertEqual(p.date("-1y 1w", basedate),
                         basedate + relativedelta(years=-1, weeks=-1))
        
        self.assertEqual(p.date("+1y 2d 5h 12s", basedate),
                         basedate + relativedelta(years=1, days=2, hours=5, seconds=12))
        
    def test_bundle_subs(self):
        p = english.bundle
        
        self.test_time(p)
        self.test_date(p)
        self.test_plustime(p)
        self.test_dayname(p)
        self.test_reldate(p)
        
    def test_bundle(self):
        p = english.bundle
        
        self.assert_atime(p.date("mar 29 1972 2:45am", basedate),
                          year=1972, month=3, day=29, hour=2, minute=45)
        self.assert_atime(p.date("16:10:45 14 February 2005", basedate),
                          year=2005, month=2, day=14, hour=16, minute=10, second=45)
        self.assert_atime(p.date("1985 sept 12 12:01", basedate),
                          year=1985, month=9, day=12, hour=12, minute=1)
        self.assert_atime(p.date("5pm 21st oct 2005", basedate),
                          year=2005, month=10, day=21, hour=17)
        self.assert_atime(p.date("5:59:59pm next thur", basedate),
                          year=2010, month=9, day=23, hour=17, minute=59, second=59)
    
    def test_ranges(self):
        p = english.torange
        
        self.assert_timespan(p.date("last tuesday to next tuesday", basedate),
                             dict(year=2010, month=9, day=14),
                             dict(year=2010, month=9, day=21))
        self.assert_timespan(p.date("last monday to dec 25", basedate),
                             dict(year=2010, month=9, day=13),
                             dict(year=2010, month=12, day=25))
        self.assert_timespan(p.date("oct 25 to feb 14", basedate),
                             dict(year=2009, month=10, day=25),
                             dict(year=2010, month=2, day=14))
        self.assert_timespan(p.date("oct 25 2009 to feb 14 2008", basedate),
                             dict(year=2008, month=2, day=14),
                             dict(year=2009, month=10, day=25))
        self.assert_timespan(p.date("3am oct 12 to 5pm", basedate),
                             dict(year=2010, month=10, day=12, hour=3),
                             dict(year=2010, month=10, day=12, hour=17))
        self.assert_timespan(p.date("3am feb 12 to 5pm today", basedate),
                             dict(year=2010, month=2, day=12, hour=3),
                             dict(year=2010, month=9, day=20, hour=17))
        self.assert_timespan(p.date("feb to oct", basedate),
                             dict(year=2010, month=2, day=1),
                             dict(year=2010, month=10, day=31))
        self.assert_timespan(p.date("oct 25 2005 11am to 5pm tomorrow", basedate),
                             dict(year=2005, month=10, day=25, hour=11),
                             dict(year=2010, month=9, day=21, hour=17))
        self.assert_timespan(p.date("oct 5 2005 to november 20", basedate),
                             dict(year=2005, month=10, day=5),
                             dict(year=2005, month=11, day=20))
        
        self.assert_datespan(p.date("-2d to +1w", basedate),
                             basedate + relativedelta(days=-2),
                             basedate + relativedelta(weeks=1))
        


if __name__ == '__main__':
    unittest.main()






import unittest

import os, os.path, threading, time
from shutil import rmtree

from whoosh.filedb.filestore import FileStorage
from whoosh.support.filelock import try_for
from whoosh.util import length_to_byte, byte_to_length


class TestMisc(unittest.TestCase):
    def make_storage(self, name):
        if not os.path.exists(name):
            os.mkdir(name)
        st = FileStorage(name)
        return st
    
    def destroy_dir(self, name):
        try:
            rmtree(name)
        except:
            raise
    
    def clean_file(self, path):
        if os.path.exists(path):
            try:
                os.remove(path)
            except:
                raise
    
    def test_filelock_simple(self):
        st = self.make_storage("testindex")
        lock1 = st.lock("testlock")
        lock2 = st.lock("testlock")
        self.assertFalse(lock1 is lock2)
        
        self.assertTrue(lock1.acquire())
        self.assertTrue(os.path.exists("testindex/testlock"))
        self.assertFalse(lock2.acquire())
        lock1.release()
        self.assertTrue(lock2.acquire())
        self.assertFalse(lock1.acquire())
        lock2.release()
        
        self.destroy_dir("testindex")
    
    def test_threaded_filelock(self):
        st = self.make_storage("testindex")
        lock1 = st.lock("testlock")
        result = []
        
        # The thread function tries to acquire the lock and then quits
        def fn():
            lock2 = st.lock("testlock")
            gotit = try_for(lock2.acquire, 1.0, 0.1)
            if gotit:
                result.append(True)
                lock2.release()
        t = threading.Thread(target=fn)
        
        # Acquire the lock in this thread
        lock1.acquire()
        # Start the other thread trying to acquire the lock
        t.start()
        # Wait for a bit
        time.sleep(0.15)
        # Release the lock
        lock1.release()
        # Wait for the other thread to finish
        t.join()
        # If the other thread got the lock, it should have appended True to the
        # "results" list.
        self.assertEqual(result, [True])
        
        self.destroy_dir("testindex")
        
    def test_length_byte(self):
        source = range(11)
        xform = [length_to_byte(n) for n in source]
        result = [byte_to_length(n) for n in xform]
        self.assertEqual(source, result)


if __name__ == '__main__':
    unittest.main()



import unittest
import os.path, shutil

from whoosh import fields, index


class TestSchema(unittest.TestCase):
    def create_index(self, schema, ixname, dir="testindex"):
        if not os.path.exists(dir):
            os.mkdir(dir)
        return index.create_in(dir, schema, indexname=ixname)
        
    def destroy_index(self, dir="testindex"):
        try:
            if os.path.exists(dir):
                shutil.rmtree(dir)
        except OSError:
            pass
            
    def test_addfield(self):
        schema = fields.Schema(id=fields.ID(stored=True), content=fields.TEXT)
        ix = self.create_index(schema, "addfield")
        
        try:
            w = ix.writer()
            w.add_document(id=u"a", content=u"alfa")
            w.add_document(id=u"b", content=u"bravo")
            w.add_document(id=u"c", content=u"charlie")
            w.commit()
            
            ix.add_field("added", fields.KEYWORD(stored=True))
            
            w = ix.writer()
            w.add_document(id=u"d", content=u"delta", added=u"fourth")
            w.add_document(id=u"e", content=u"echo", added=u"fifth")
            w.commit(merge=False)
            
            s = ix.searcher()
            try:
                self.assertEqual(s.document(id=u"d"), {"id": "d", "added": "fourth"})
                self.assertEqual(s.document(id=u"b"), {"id": "b"})
            finally:
                s.close()
        finally:
            ix.close()
            self.destroy_index()
    
    def test_removefield(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               content=fields.TEXT,
                               city=fields.KEYWORD(stored=True))
        ix = self.create_index(schema, "removefield")
        
        try:
            w = ix.writer()
            w.add_document(id=u"b", content=u"bravo", city=u"baghdad")
            w.add_document(id=u"c", content=u"charlie", city=u"cairo")
            w.add_document(id=u"d", content=u"delta", city=u"dakar")
            w.commit()
            
            s = ix.searcher()
            try:
                self.assertEqual(s.document(id=u"c"),
                                 {"id": "c", "city": "cairo"})
            finally:
                s.close()
            
            w = ix.writer()
            w.remove_field("content")
            w.remove_field("city")
            w.commit()

            ixschema = ix._current_schema()
            self.assertEqual(ixschema.names(), ["id"])
            self.assertEqual(ixschema.stored_names(), ["id"])
            
            s = ix.searcher()
            self.assertFalse(("content", u"charlie") in s.reader())
            self.assertEqual(s.document(id=u"c"), {"id": u"c"})
            s.close()
        finally:
            ix.close()
            self.destroy_index()
    
    def test_optimize_away(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               content=fields.TEXT,
                               city=fields.KEYWORD(stored=True))
        ix = self.create_index(schema, "optimize")
        
        try:
            w = ix.writer()
            w.add_document(id=u"b", content=u"bravo", city=u"baghdad")
            w.add_document(id=u"c", content=u"charlie", city=u"cairo")
            w.add_document(id=u"d", content=u"delta", city=u"dakar")
            w.commit()
            
            s = ix.searcher()
            try:
                self.assertEqual(s.document(id=u"c"),
                                 {"id": "c", "city": "cairo"})
            finally:
                s.close()
            
            w = ix.writer()
            w.remove_field("content")
            w.remove_field("city")
            w.commit(optimize=True)
            
            s = ix.searcher()
            self.assertFalse(("content", u"charlie") in s.reader())
            self.assertEqual(s.document(id=u"c"), {"id": u"c"})
            s.close()
        finally:
            ix.close()
            self.destroy_index()
        

if __name__ == '__main__':
    unittest.main()





#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import re, calendar, copy
from datetime import date, time, datetime, timedelta

from whoosh.support.relativedelta import relativedelta


def rcompile(pattern):
    return re.compile(pattern, re.IGNORECASE)


def relative_days(current_wday, wday, dir):
    # Where current_wday and wday are weekday numbers i.e. 0 = monday, 1 =
    # tuesday, 2 = wednesday, etc. and dir is -1 for the past (i.e. "last x")
    # or 1 for the future ("next x")
    
    if current_wday == wday:
        return 7 * dir
    
    if dir == 1:
        return (wday + 7 - current_wday) % 7
    else:
        return (current_wday + 7 - wday) % 7 * -1


def print_debug(level, msg, *args):
    if level > 0: print ("  " * (level-1)) + (msg % args)


class atime(object):
    units = frozenset(("year", "month", "day", "hour", "minute", "second", "microsecond"))
    
    def __init__(self, year=None, month=None, day=None, hour=None, minute=None,
                 second=None, microsecond=None):
        if isinstance(year, datetime):
            self.year, self.month, self.day = year.year, year.month, year.day
            self.hour, self.minute, self.second = year.hour, year.minute, year.second
            self.microsecond = year.microsecond
        else:
            self.year, self.month, self.day = year, month, day
            self.hour, self.minute, self.second = hour, minute, second
            self.microsecond = microsecond
    
    def tuple(self):
        return (self.year, self.month, self.day, self.hour, self.minute,
                self.second, self.microsecond)
    
    def __repr__(self):
        return "%s%r" % (self.__class__.__name__, self.tuple())
    
    def date(self):
        return date(self.year, self.month, self.day)
    
    def copy(self):
        return atime(year=self.year, month=self.month, day=self.day,
                     hour=self.hour, minute=self.minute, second=self.second,
                     microsecond=self.microsecond)
    
    def replace(self, **kwargs):
        newatime = self.copy()
        for key, value in kwargs.iteritems():
            if key in self.units:
                setattr(newatime, key, value)
            else:
                raise KeyError("Unknown argument %r" % key)
        return newatime


def fill_in(at, basedate, units=atime.units):
    if isinstance(at, datetime):
        return at
    
    args = {}
    for unit in units:
        v = getattr(at, unit)
        if v is None:
            v = getattr(basedate, unit)
        args[unit] = v
    return fix(atime(**args))
    
def has_no_date(at):
    if isinstance(at, datetime):
        return False
    return at.year is None and at.month is None and at.day is None

def has_no_time(at):
    if isinstance(at, datetime):
        return False
    return at.hour is None and at.minute is None and at.second is None and at.microsecond is None

def is_ambiguous(at):
    if isinstance(at, datetime):
        return False
    return any((getattr(at, attr) is None) for attr in atime.units)

def is_void(at):
    if isinstance(at, datetime):
        return False
    return all((getattr(at, attr) is None) for attr in atime.units)

def fix(at):
    if is_ambiguous(at) or isinstance(at, datetime):
        return at
    return datetime(year=at.year, month=at.month, day=at.day, hour=at.hour,
                    minute=at.minute, second=at.second, microsecond=at.microsecond)

def start_of_year(dt):
    return dt.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
def start_of_month(dt):
    return dt.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
def start_of_day(dt):
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)
def start_of_hour(dt):
    return dt.replace(minute=0, second=0, microsecond=0)
def start_of_minute(dt):
    return dt.replace(second=0, microsecond=0)

def end_of_year(dt):
    lastday = calendar.monthrange(dt.year, dt.month)[1]
    return dt.replace(month=12, day=lastday, hour=12, minute=59, second=59, microsecond=9999999)
def end_of_month(dt):
    lastday = calendar.monthrange(dt.year, dt.month)[1]
    return dt.replace(month=12, day=lastday, hour=12, minute=59, second=59, microsecond=9999999)
def end_of_day(dt):
    return dt.replace(hour=12, minute=59, second=59, microsecond=9999999)
def end_of_hour(dt):
    return dt.replace(minute=59, second=59, microsecond=9999999)
def end_of_minute(dt):
    return dt.replace(second=59, microsecond=9999999)


class timespan(object):
    def __init__(self, start, end, basedate):
        start = copy.copy(start)
        end = copy.copy(end)
        year_was_amb = start.year is None
        
        if has_no_date(start) and has_no_date(end):
            start = start.replace(year=basedate.year, month=basedate.month, day=basedate.day)
            end = end.replace(year=basedate.year, month=basedate.month, day=basedate.day)
        else:
            if start.year is None and end.year is None:
                start.year = end.year = basedate.year
            elif start.year is None:
                start.year = end.year
            elif end.year is None:
                end.year = start.year
            
            if start.month is None and end.month is None:
                start.month = 1
                end.month = 12
            elif start.month is None:
                start.month = end.month
            elif end.month is None:
                end.month = start.month
                
            if start.day is None and end.day is None:
                start.day = 1
                end.day = 31
            elif start.day is None:
                start.day = end.day
            elif end.day is None:
                end.day = start.day
            
            if start.date() > end.date():
                if year_was_amb:
                    start = start.replace(year=start.year-1)
                else:
                    start, end = end, start
        
        self.start = start
        self.end = end
        
    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.start, self.end)


class Props(object):
    def __init__(self, **args):
        self.__dict__ = args
    
    def __repr__(self):
        return repr(self.__dict__)
    
    def get(self, key, default=None):
        return self.__dict__.get(key, default)


class ParserBase(object):
    def to_parser(self, e):
        if isinstance(e, basestring):
            return Regex(e)
        else:
            return e
    
    def parse(self, text, dt, pos=0, debug=-9999):
        raise NotImplementedError
    
    def date(self, text, dt=None, pos=0, debug=-9999):
        if dt is None:
            dt = datetime.now()
        
        d, pos = self.parse(text, dt, pos, debug + 1)
        if isinstance(d, atime):
            d = fix(d)
        
        return d


class MultiBase(ParserBase):
    def __init__(self, elements, name=None):
        self.elements = [self.to_parser(e) for e in elements]
        self.name = name
        
    def __repr__(self):
        return "%s<%s>%r" % (self.__class__.__name__, self.name or '', self.elements)


class Sequence(MultiBase):
    def __init__(self, elements, sep="(\\s+|\\s*,\\s*)", name=None):
        super(Sequence, self).__init__(elements, name)
        if sep:
            self.sep_expr = rcompile(sep)
        else:
            self.sep_expr = None
    
    def parse(self, text, dt, pos=0, debug=-9999):
        d = atime()
        first = True
        
        print_debug(debug, "Seq %s sep=%r text=%r", self.name, self.sep_expr.pattern, text[pos:])
        for e in self.elements:
            print_debug(debug, "Seq %s text=%r", self.name, text[pos:])
            if self.sep_expr and not first:
                print_debug(debug, "Seq %s looking for sep", self.name)
                m = self.sep_expr.match(text, pos)
                if m:
                    pos = m.end()
                else:
                    print_debug(debug, "Seq %s didn't find sep", self.name)
                    return (None, None)
            
            print_debug(debug, "Seq %s trying=%r", self.name, e)
            at, pos = e.parse(text, dt, pos, debug + 1)
            print_debug(debug, "Seq %s result=%r", self.name, at)
            if not at:
                return (None, None)
            d = fill_in(d, at)
            
            first = False
        
        print_debug(debug, "Seq %s final=%r", self.name, d)
        return (d, pos)


class Combo(Sequence):
    def __init__(self, elements, fn=None, sep="(\\s+|\\s*,\\s*)", min=2, max=2,
                 name=None):
        super(Combo, self).__init__(elements, sep=sep, name=name)
        self.fn = fn
        self.min = min
        self.max = max
    
    def parse(self, text, dt, pos=0, debug=-9999):
        dates = []
        first = True
        
        print_debug(debug, "Combo %s sep=%r text=%r", self.name, self.sep_expr.pattern, text[pos:])
        for e in self.elements:
            if self.sep_expr and not first:
                print_debug(debug, "Combo %s looking for sep at %r", self.name, text[pos:])
                m = self.sep_expr.match(text, pos)
                if m:
                    pos = m.end()
                else:
                    print_debug(debug, "Combo %s didn't find sep", self.name)
                    return (None, None)
            
            print_debug(debug, "Combo %s trying=%r", self.name, e)
            at, pos = e.parse(text, dt, pos, debug + 1)
            print_debug(debug, "Combo %s result=%r", self.name, at)
            if at is None:
                return (None, None)
            
            first = False
            if is_void(at):
                continue
            if len(dates) == self.max:
                print_debug(debug, "Combo %s length > %s", self.name, self.max)
                return (None, None)
            dates.append(at)
        
        print_debug(debug, "Combo %s dates=%r", self.name, dates)
        if len(dates) < self.min:
            print_debug(debug, "Combo %s length < %s", self.name, self.min)
            return (None, None)
        
        return (self.dates_to_timespan(dates, dt), pos)
    
    def dates_to_timespan(self, dates, basedate):
        if self.fn:
            return self.fn(dates, basedate)
        elif len(dates) == 2:
            return timespan(dates[0], dates[1], basedate)
        else:
            raise Exception("Don't know what to do with %r" % (dates, ))


class Choice(MultiBase):
    def parse(self, text, dt, pos=0, debug=-9999):
        print_debug(debug, "Choice %s text=%r", self.name, text[pos:])
        for e in self.elements:
            print_debug(debug, "Choice %s trying=%r", self.name, e)
            d, newpos = e.parse(text, dt, pos, debug + 1)
            if d:
                print_debug(debug, "Choice %s matched", self.name)
                return (d, newpos)
        print_debug(debug, "Choice %s no match", self.name)
        return (None, None)


class Bag(MultiBase):
    def __init__(self, elements, sep="(\\s+|\\s*,\\s*)", onceper=True,
                 requireall=False, allof=None, anyof=None, name=None):
        super(Bag, self).__init__(elements, name)
        self.sep_expr = rcompile(sep)
        self.onceper = onceper
        self.requireall = requireall
        self.allof = allof
        self.anyof = anyof
    
    def parse(self, text, dt, pos=0, debug=-9999):
        first = True
        d = atime()
        seen = [False] * len(self.elements)
        
        while True:
            newpos = pos
            print_debug(debug, "Bag %s text=%r", self.name, text[pos:])
            if not first:
                print_debug(debug, "Bag %s looking for sep", self.name)
                m = self.sep_expr.match(text, pos)
                if m:
                    newpos = m.end()
                else:
                    print_debug(debug, "Bag %s didn't find sep", self.name)
                    break
            
            for i, e in enumerate(self.elements):
                print_debug(debug, "Bag %s trying=%r", self.name, e)
                at, xpos  = e.parse(text, dt, newpos, debug + 1)
                print_debug(debug, "Bag %s result=%r", self.name, at)
                if at:
                    if self.onceper and seen[i]:
                        return (None, None)
                    
                    d = fill_in(d, at)
                    newpos = xpos
                    seen[i] = True
                    break
            else:
                break
            
            pos = newpos
            if self.onceper and all(seen):
                break
            
            first = False
        
        if (not any(seen)
            or (self.allof and not all(seen[pos] for pos in self.allof))
            or (self.anyof and not any(seen[pos] for pos in self.anyof))
            or (self.requireall and not all(seen))):
            return (None, None)
        
        print_debug(debug, "Bag %s final=%r", self.name, d)
        return (d, pos)
    

class Optional(ParserBase):
    def __init__(self, element):
        self.element = self.to_parser(element)
    
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.element)
    
    def parse(self, text, dt, pos=0, debug=-9999):
        d, pos = self.element.parse(text, dt, pos, debug + 1)
        if d:
            return (d, pos)
        else:
            return (atime(), pos)


class Regex(ParserBase):
    fn = None
    
    def __init__(self, pattern, fn=None, modify=None):
        self.pattern = pattern
        self.expr = rcompile(pattern)
        self.fn = fn
        self.modify = modify
    
    def __repr__(self):
        return "<%r>" % (self.pattern, )
    
    def parse(self, text, dt, pos=0, debug=-9999):
        m = self.expr.match(text, pos)
        if not m:
            return (None, None)
        
        props = self.extract(m)
        self.modify_props(props)
        d = self.props_to_date(props, dt)
        if d:
            return (d, m.end())
        else:
            return (None, None)
    
    def extract(self, match):
        d = match.groupdict()
        for key, value in d.iteritems():
            try:
                value = int(value)
                d[key] = value
            except (ValueError, TypeError):
                pass
        return Props(**d)
    
    def modify_props(self, props):
        if self.modify:
            self.modify(props)
            
    def props_to_date(self, props, dt):
        if self.fn:
            return self.fn(props, dt)
        else:
            args = {}
            for key in atime.units:
                args[key] = props.get(key)
            return atime(**args)

    
class Month(Regex):
    def __init__(self, *patterns):
        self.patterns = patterns
        self.exprs = [rcompile(pat) for pat in self.patterns]
        
        self.pattern = ("(?P<month>"
                        + "|".join("(%s)" % pat for pat in self.patterns)
                        + ")")
        self.expr = rcompile(self.pattern)
        
    def modify_props(self, p):
        text = p.month
        for i, expr in enumerate(self.exprs):
            m = expr.match(text)
            if m:
                p.month = i + 1
                break
            

class Delta(Regex):
    def __init__(self, pattern, **args):
        super(Delta, self).__init__(pattern)
        self.args = args
    
    def props_to_date(self, p, dt):
        args = {}
        dt = dt.replace(dt.year + p.get("years", self.args.get("years", 0)))
        for key in ("weeks", "days", "hours", "minutes", "seconds"):
            args[key] = p.get(key, self.args.get(key, 0))
        return dt + timedelta(**args)


class DateParser(object):
    day = Regex("(?P<day>([123][0-9])|[1-9])(?=(\\W|$))(?!=:)",
                lambda p, dt: atime(day=p.day))
    year = Regex("(?P<year>[0-9]{4})(?=(\\W|$))",
                 lambda p, dt: atime(year=p.year))
    time24 = Regex("(?P<hour>([0-1][0-9])|(2[0-3])):(?P<mins>[0-5][0-9])(:(?P<secs>[0-5][0-9])(\\.(?P<usecs>[0-9]{1,5}))?)?(?=(\\W|$))",
                   lambda p, dt: atime(hour=p.hour, minute=p.mins, second=p.secs,
                                       microsecond=p.usecs))
    def __init__(self):
        self.setup()
        

class English(DateParser):
    day = Regex("(?P<day>([123][0-9])|[1-9])(st|nd|rd|th)?(?=(\\W|$))",
                lambda p, dt: atime(day=p.day))
    
    def setup(self):
        self.time12 = Regex("(?P<hour>[1-9]|11|12)(:(?P<mins>[0-5][0-9])(:(?P<secs>[0-5][0-9])(\\.(?P<usecs>[0-9]{1,5}))?)?)?\\s*(?P<ampm>am|pm)(?=(\\W|$))",
                            self.modify_time12_props)
        
        rel_hours = "((?P<hours>[0-9]+) *(hours|hour|hrs|hr|hs|h))?"
        rel_mins = "((?P<mins>[0-9]+) *(minutes|minute|mins|min|ms|m))?"
        rel_secs = "((?P<secs>[0-9]+) *(seconds|second|secs|sec|s))?"
        self.plustime = Regex("(?P<dir>[+-]) *%s *%s *%s(?=(\\W|$))" % (rel_hours, rel_mins, rel_secs),
                              self.plustime_to_date)
        
        midnight = Regex("midnight", lambda p, dt: atime(hour=0, minute=0, second=0, microsecond=0))
        noon = Regex("noon", lambda p, dt: atime(hour=12, minute=0, second=0, microsecond=0))
        now = Delta("now")
        self.time = Choice((self.time12, self.time24, midnight, noon, now), name="time")
        
        tomorrow = Regex("tomorrow", self.tomorrow_to_date)
        yesterday = Regex("yesterday", self.yesterday_to_date)
        thisyear = Regex("this year", lambda p, dt: atime(year=dt.year))
        thismonth = Regex("this month", lambda p, dt: atime(year=dt.year, month=dt.month))
        today = Regex("today", lambda p, dt: atime(year=dt.year, month=dt.month, day=dt.day))
        
        rel_years = "((?P<years>[0-9]+) *(years|year|yrs|yr|ys|y))?"
        rel_months = "((?P<months>[0-9]+) *(months|month|mons|mon|mos|mo))?"
        rel_weeks = "((?P<weeks>[0-9]+) *(weeks|week|wks|wk|ws|w))?"
        rel_days = "((?P<days>[0-9]+) *(days|day|dys|dy|ds|d))?"
        self.plusdate = Regex("(?P<dir>[+-]) *%s *%s *%s *%s *%s *%s *%s(?=(\\W|$))" % (rel_years, rel_months, rel_weeks, rel_days, rel_hours, rel_mins, rel_secs),
                              self.plusdate_to_date)
        
        daynames = ("monday|mon|mo", "tuesday|tues|tue|tu", "wednesday|wed|we",
                    "thursday|thur|thu|th", "friday|fri|fr", "saturday|sat|sa",
                    "sunday|sun|su")
        self.dayname_exprs = tuple(rcompile(pat) for pat in daynames)
        self.dayname = Regex("(?P<dir>last|next) +(?P<day>%s)(?=(\\W|$))" % ("|".join(daynames)),
                             self.dayname_to_date)
        
        self.month = Month("january|jan", "february|febuary|feb", "march|mar",
                           "april|apr", "may", "june|jun", "july|jul", "august|aug",
                           "september|sept|sep", "october|oct", "november|nov",
                           "december|dec")
        
        # If you specify a day number you must also specify a year and/or a
        # month... this Choice captures that constraint
        
        self.date = Choice((Sequence((self.day, self.month, self.year), name="dmy"),
                            Sequence((self.month, self.day, self.year), name="mdy"),
                            Sequence((self.year, self.month, self.day), name="ymd"),
                            Sequence((self.year, self.day, self.month), name="ydm"),
                            Sequence((self.day, self.month), name="dm"),
                            Sequence((self.month, self.day), name="md"),
                            Sequence((self.month, self.year), name="my"),
                            self.month, self.year, self.dayname, tomorrow,
                            yesterday, thisyear, thismonth, today, now,
                            ), name="date")
        
        self.datetime = Bag((self.time, self.date), name="datetime")
        self.bundle = Choice((self.plusdate, self.datetime), name="bundle")
        
        self.torange = Combo((self.bundle, "to", self.bundle), name="torange")
    
    def plusdate_to_date(self, p, dt):
        if p.dir == "-":
            dir = -1
        else:
            dir = 1
        delta = relativedelta(years=(p.get("years") or 0) * dir,
                              months=(p.get("months") or 0) * dir,
                              weeks=(p.get("weeks") or 0) * dir,
                              days=(p.get("days") or 0) * dir,
                              hours=(p.get("hours") or 0) * dir,
                              minutes=(p.get("mins") or 0) * dir,
                              seconds=(p.get("secs") or 0) * dir)
        return dt + delta
    
    def plustime_to_date(self, p, dt):
            if p.dir == "-":
                dir = -1
            else:
                dir = 1
            delta = timedelta(hours=(p.get("hours") or 0) * dir,
                              minutes=(p.get("mins") or 0) * dir,
                              seconds=(p.get("secs") or 0) * dir)
            return dt + delta 
    
    def modify_time12_props(self, p, dt):
        if p.hour == 12:
            if p.ampm == "am":
                hr = 0
            else:
                hr = 12
        else:
            hr = p.hour
            if p.ampm == "pm":
                hr += 12
        return atime(hour=hr, minute=p.mins, second=p.secs, microsecond=p.usecs)
    
    def tomorrow_to_date(self, p, dt):
        d = dt.date() + timedelta(days=+1)
        return atime(year=d.year, month=d.month, day=d.day)
    
    def yesterday_to_date(self, p, dt):
        d = dt.date() + timedelta(days=-1)
        return atime(year=d.year, month=d.month, day=d.day)
        
    def dayname_to_date(self, p, dt):
        if p.dir == "last":
            dir = -1
        else:
            dir = 1
        
        for daynum, expr in enumerate(self.dayname_exprs):
            m = expr.match(p.day)
            if m:
                break
        current_daynum = dt.weekday()
        days_delta = relative_days(current_daynum, daynum, dir)
        
        d = dt.date() + timedelta(days=days_delta)
        return atime(year=d.year, month=d.month, day=d.day)
    
    
    
    
    
    
    
    
    
    






#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from threading import Lock

from whoosh.filedb.filepostings import FilePostingReader
from whoosh.filedb.filetables import (TermIndexReader, StoredFieldReader,
                                      LengthReader, TermVectorReader)
from whoosh.matching import ExcludeMatcher
from whoosh.reading import IndexReader, TermNotFound
from whoosh.util import protected


# Reader class

class SegmentReader(IndexReader):
    def __init__(self, storage, schema, segment, generation=None):
        self.storage = storage
        self.schema = schema
        self.segment = segment
        self._generation = generation
        
        # Term index
        tf = storage.open_file(segment.termsindex_filename)
        self.termsindex = TermIndexReader(tf)
        
        # Term postings file, vector index, and vector postings: lazy load
        self.postfile = None
        self.vectorindex = None
        self.vpostfile = None
        
        # Stored fields file
        sf = storage.open_file(segment.storedfields_filename, mapped=False)
        self.storedfields = StoredFieldReader(sf)
        
        # Field length file
        self.fieldlengths = None
        if self.schema.has_scorable_fields():
            flf = storage.open_file(segment.fieldlengths_filename)
            self.fieldlengths = LengthReader(flf, segment.doc_count_all())
        
        # Copy methods from underlying segment
        self.has_deletions = segment.has_deletions
        self.is_deleted = segment.is_deleted
        self.doc_count = segment.doc_count
        
        # Postings file
        self.postfile = self.storage.open_file(segment.termposts_filename,
                                               mapped=False)
        
        self.dc = segment.doc_count_all()
        assert self.dc == self.storedfields.length
        
        self.is_closed = False
        self._sync_lock = Lock()

    def _open_vectors(self):
        if self.vectorindex: return
        
        storage, segment = self.storage, self.segment
        
        # Vector index
        vf = storage.open_file(segment.vectorindex_filename)
        self.vectorindex = TermVectorReader(vf)
        
        # Vector postings file
        self.vpostfile = storage.open_file(segment.vectorposts_filename,
                                           mapped=False)
    
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.segment)

    @protected
    def __contains__(self, term):
        return term in self.termsindex

    def generation(self):
        return self._generation

    def close(self):
        self.storedfields.close()
        self.termsindex.close()
        if self.postfile:
            self.postfile.close()
        if self.vectorindex:
            self.vectorindex.close()
        #if self.fieldlengths:
        #    self.fieldlengths.close()
        self.is_closed = True

    def doc_count_all(self):
        return self.dc

    def field(self, fieldname):
        return self.schema[fieldname]

    def scorable(self, fieldname):
        return self.schema[fieldname].scorable
    
    def scorable_names(self):
        return self.schema.scorable_names()
    
    def vector_names(self):
        return self.schema.vector_names()
    
    def format(self, fieldname):
        return self.schema[fieldname].format
    
    def vector_format(self, fieldname):
        return self.schema[fieldname].vector

    @protected
    def stored_fields(self, docnum):
        schema = self.schema
        return dict(item for item
                    in self.storedfields[docnum].iteritems()
                    if item[0] in schema)

    @protected
    def all_stored_fields(self):
        is_deleted = self.segment.is_deleted
        sf = self.stored_fields
        for docnum in xrange(self.segment.doc_count_all()):
            if not is_deleted(docnum):
                yield sf(docnum)

    def field_length(self, fieldname):
        return self.segment.field_length(fieldname)

    @protected
    def doc_field_length(self, docnum, fieldname, default=0):
        if self.fieldlengths is None: return default
        return self.fieldlengths.get(docnum, fieldname, default=default)

    def max_field_length(self, fieldname):
        return self.segment.max_field_length(fieldname)

    @protected
    def has_vector(self, docnum, fieldname):
        self._open_vectors()
        return (docnum, fieldname) in self.vectorindex

    @protected
    def __iter__(self):
        schema = self.schema
        for (fieldname, t), (totalfreq, _, postcount) in self.termsindex:
            if fieldname not in schema:
                continue
            yield (fieldname, t, postcount, totalfreq)

    def _test_field(self, fieldname):
        if fieldname not in self.schema:
            raise TermNotFound("No field %r" % fieldname)
        if self.schema[fieldname].format is None:
            raise TermNotFound("Field %r is not indexed" % fieldname)

    @protected
    def iter_from(self, fieldname, text):
        schema = self.schema
        self._test_field(fieldname)
        for (fn, t), (totalfreq, _, postcount) in self.termsindex.items_from((fieldname, text)):
            if fn not in schema:
                continue
            yield (fn, t, postcount, totalfreq)

    @protected
    def _term_info(self, fieldname, text):
        self._test_field(fieldname)
        try:
            return self.termsindex[(fieldname, text)]
        except KeyError:
            raise TermNotFound("%s:%r" % (fieldname, text))

    def doc_frequency(self, fieldname, text):
        self._test_field(fieldname)
        try:
            return self._term_info(fieldname, text)[2]
        except TermNotFound:
            return 0

    def frequency(self, fieldname, text):
        self._test_field(fieldname)
        try:
            return self._term_info(fieldname, text)[0]
        except TermNotFound:
            return 0

    def lexicon(self, fieldname):
        # The base class has a lexicon() implementation that uses iter_from()
        # and throws away the value, but overriding to use
        # FileTableReader.keys_from() is much, much faster.

        self._test_field(fieldname)
        return self.expand_prefix(fieldname, '')

    @protected
    def expand_prefix(self, fieldname, prefix):
        # The base class has an expand_prefix() implementation that uses
        # iter_from() and throws away the value, but overriding to use
        # FileTableReader.keys_from() is much, much faster.

        self._test_field(fieldname)
        for fn, t in self.termsindex.keys_from((fieldname, prefix)):
            if fn != fieldname or not t.startswith(prefix):
                return
            yield t

    def postings(self, fieldname, text, exclude_docs=frozenset(), scorer=None):
        self._test_field(fieldname)
        format = self.format(fieldname)
        try:
            offset = self.termsindex[(fieldname, text)][1]
        except KeyError:
            raise TermNotFound("%s:%r" % (fieldname, text))

        if self.segment.deleted and exclude_docs:
            exclude_docs = self.segment.deleted | exclude_docs
        elif self.segment.deleted:
            exclude_docs = self.segment.deleted

        postreader = FilePostingReader(self.postfile, offset, format,
                                       scorer=scorer, fieldname=fieldname,
                                       text=text)
        if exclude_docs:
            postreader = ExcludeMatcher(postreader, exclude_docs)
            
        return postreader
    
    def vector(self, docnum, fieldname):
        if fieldname not in self.schema:
            raise TermNotFound("No  field %r" % fieldname)
        vformat = self.vector_format(fieldname)
        if not vformat:
            raise Exception("No vectors are stored for field %r" % fieldname)
        
        self._open_vectors()
        offset = self.vectorindex.get((docnum, fieldname))
        if offset is None:
            raise Exception("No vector found"
                            " for document %s field %r" % (docnum, fieldname))
        
        return FilePostingReader(self.vpostfile, offset, vformat, stringids=True)

        















